/* eslint-env es6 */
/* global exports module define require */
/* eslint new-cap: 0, no-multi-spaces: 0 */
/**
 * Converts SmartType to a JavaScript function
 */
(function () {
    "use strict";

    var factory = function factory(smarttype, simplifier, P) {
        var parser = smarttype.parser;
        var Simplifier = simplifier.Simplifier;

        var UnsupportedMathError = function UnsupportedMathError(token) {
            this.name = "UnsupportedMathError";
            this.message = "Unsupported math: " + token;
        };
        UnsupportedMathError.prototype = new Error();
        UnsupportedMathError.prototype.constructor = UnsupportedMathError;

        var Visitor = function Visitor(options) {
            this.options = options;
        };
        var add_f = function add_f(operands, bindings) {
            var r = operands[0](bindings);
            for (var i = 1; i < operands.length - 1; i += 2) {
                if (operands[i] === "+") {
                    r += operands[i + 1](bindings);
                } else {
                    r -= operands[i + 1](bindings);
                }
            }
            return r;
        };
        Visitor.prototype.visit_addition = function () {
            return add_f.bind(undefined, Array.prototype.slice.call(arguments, 1));
        };
        Visitor.prototype.visit_addition_sign = function (sign) {
            if (sign[0] === "\\") {
                throw new UnsupportedMathError(sign);
            }
            return sign;
        };
        var multiply_f = function multiply_f(operands, bindings) {
            var r = 1;
            for (var i = 0; i < operands.length; i += 2) {
                r *= operands[i](bindings);
            }
            return r;
        };
        Visitor.prototype.visit_multiplication = function () {
            return multiply_f.bind(undefined, Array.prototype.slice.call(arguments, 1));
        };
        Visitor.prototype.visit_multiplication_op = function (visible) {
            return;
        };

        var reciprocal = function reciprocal(x) {
            return 1 / x;
        };
        var div = function div(x, y) {
            return x / y;
        };
        var id = function id(n) {
            return n;
        };
        var neg = function neg(x) {
            return -x;
        };

        var const_f = function const_f(c, bindings) {
            return c;
        };
        var apply1_f = function apply1_f(f, operand, bindings) {
            return f.call(undefined, operand(bindings));
        };
        var apply2_f = function apply2_f(f, op1, op2, bindings) {
            return f.call(undefined, op1(bindings), op2(bindings));
        };
        var log_f = function log_f(base, operand, bindings) {
            return Math.log(operand) / Math.log(base(bindings));
        };

        Visitor.prototype.visit_binop = new P.Function().when(P._, P("left"), "root", P("right"), function (_) {
            return apply2_f.bind(undefined, Math.pow, _.right, apply1_f.bind(undefined, reciprocal, _.left));
        }).when(P._, P("left"), "^", P("right"), function (_) {
            return apply2_f.bind(undefined, Math.pow, _.left, _.right);
        }).when(P._, P("left"), "/", P("right"), function (_) {
            return apply2_f.bind(undefined, div, _.left, _.right);
        }).when(P._, P("left"), "=", P("right"), function (_) {
            return { left: _.left, op: "=", right: _.right };
        }).when(P._, "log", "_", P("base"), function (_) {
            return log_f.bind(undefined, _.base);
        }).when(P._, P("left", P.isA(Array)), ",", P("right"), function (_) {
            _.left.push(_.right);return _.left;
        }).when(P._, P("left"), ",", P("right"), function (_) {
            return [_.left, _.right];
        }).when(P._, P._, P("op"), P._, function (_) {
            throw new UnsupportedMathError(_.op);
        }).end;
        Visitor.prototype.visit_funcname = function (name) {
            return name;
        };
        var sum_f = function sum_f(variable, start, end, val, bindings) {
            var newbindings = {};
            if (bindings) {
                for (var k in bindings) {
                    newbindings[k] = bindings[k];
                }
            }
            variable = variable(variable_f);
            start = start(bindings);
            end = end(bindings);
            var r = 0;
            for (var i = start; i <= end; i++) {
                newbindings[variable] = i;
                r += val(newbindings);
            }
            return r;
        };
        var prod_f = function prod_f(variable, start, end, val, bindings) {
            var newbindings = {};
            if (bindings) {
                for (var k in bindings) {
                    newbindings[k] = bindings[k];
                }
            }
            variable = variable(variable_f);
            start = start(bindings);
            end = end(bindings);
            var r = 1;
            for (var i = start; i <= end; i++) {
                newbindings[variable] = i;
                r *= val(newbindings);
            }
            return r;
        };
        Visitor.prototype.visit_bigop = new P.Function().when(P._, "sum", { left: P("variable"),
            op: "=",
            right: P("start") }, P("end"), function (_) {
            return { func: "sum", variable: _.variable, start: _.start, end: _.end };
        }).when(P._, "prod", { left: P("variable"),
            op: "=",
            right: P("start") }, P("end"), function (_) {
            return { func: "prod", variable: _.variable, start: _.start, end: _.end };
        }).when(P._, P("op"), P("sub"), P("sup"), function (_) {
            throw new UnsupportedMathError(_.op);
        }).end;
        Visitor.prototype.visit_sign = new P.Function().when(P._, "+", P("value"), function (_) {
            return _.value;
        }).when(P._, "-", P("value"), function (_) {
            return apply1_f.bind(undefined, neg, _.value);
        }).when(P._, P("op"), P._, function (_) {
            throw new UnsupportedMathError(_.op);
        }).end;
        var min_f = function min_f(values, bindings) {
            var v = [];
            for (var i = 0; i < values.length; i++) {
                v.push(values[i](bindings));
            }
            return Math.min.apply(undefined, v);
        };
        var max_f = function max_f(values, bindings) {
            var v = [];
            for (var i = 0; i < values.length; i++) {
                v.push(values[i](bindings));
            }
            return Math.max.apply(undefined, v);
        };
        Visitor.prototype.visit_func = new P.Function()
        // special functions
        .when(P._, { func: "sum",
            variable: P("variable"),
            start: P("start"),
            end: P("end") }, P("value"), function (_) {
            return sum_f.bind(undefined, _.variable, _.start, _.end, _.value);
        }).when(P._, { func: "prod",
            variable: P("variable"),
            start: P("start"),
            end: P("end") }, P("value"), function (_) {
            return prod_f.bind(undefined, _.variable, _.start, _.end, _.value);
        }).when(P._, P("func", P.isA(Function)), P("value"), function (_) {
            return apply1_f.bind(undefined, _.func, _.value);
        }).when(P._, "min", P("values", P.isAn(Array)), function (_) {
            return min_f.bind(undefined, _.values);
        }).when(P._, "min", P("value"), function (_) {
            return _.value;
        }).when(P._, "max", P("values", P.isAn(Array)), function (_) {
            return max_f.bind(undefined, _.values);
        }).when(P._, "max", P("value"), function (_) {
            return _.value;
        })
        // regular functions
        .when(P._, "floor", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.floor, _.value);
        }).when(P._, "ceil", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.ceil, _.value);
        }).when(P._, "sqrt", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.sqrt, _.value);
        }).when(P._, "abs", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.abs, _.value);
        }).when(P._, "ln", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.log, _.value);
        }).when(P._, "sin", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.sin, _.value);
        }).when(P._, "cos", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.cos, _.value);
        }).when(P._, "tan", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.tan, _.value);
        }).when(P._, "arcsin", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.asin, _.value);
        }).when(P._, "arccos", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.acos, _.value);
        }).when(P._, "arctan", P("value"), function (_) {
            return apply1_f.bind(undefined, Math.atan, _.value);
        }).when(P._, P("name"), P("value"), function (_) {
            throw new UnsupportedMathError(_.name);
        }).end;
        Visitor.prototype.visit_postfunc = new P.Function()
        // FIXME: support factorial
        .when(P._, P("name"), P("value"), function (_) {
            throw new UnsupportedMathError(_.name);
        }).end;
        Visitor.prototype.visit_number = function (num) {
            return id.bind(undefined, parseFloat(num.value));
        };
        var variable_f = function variable_f(name, bindings) {
            if (bindings === variable_f) {
                return name;
            }
            return bindings[name];
        };
        Visitor.prototype.visit_variable = new P.Function().when({ name: "e" }, function (_) {
            return const_f.bind(undefined, Math.E);
        })
        // FIXME: make sure variable is available
        .when({ name: P("name") }, function (_) {
            return variable_f.bind(undefined, _.name);
        }).end;
        var greek = /^(?:alpha|beta|[Gg]amma|[Dd]elta|(?:var)?epsilon|(?:z||(?:T|(?:var)?t)h)eta|iota|(?:var)?kappa|[Ll]ambda|[mn]u|[Xx]i|omicron|(?:P|(?:var)?p)i|(?:var)?rho|(?:S|(?:var)?s)igma|tau|[Uu]psilon|(?:(?:P|(?:var)?p)|c)hi|[Pp]si|[Oo]mega)$/;
        Visitor.prototype.visit_symbol = new P.Function().when({ name: "pi" }, function (_) {
            return const_f.bind(undefined, Math.PI);
        }).when({ name: P("name", greek) }, function (_) {
            return variable_f.bind(undefined, _.name);
        }).when({ name: P("name") }, function (_) {
            throw new UnsupportedMathError(_.name);
        }).end;
        Visitor.prototype.visit_bracket = new P.Function()
        // unprocessed integral
        .when(P._, "int", P("val"), P("rbkt"), P._, function (_) {
            throw new UnsupportedMathError("int");
        })
        // already processed integral
        .when(P._, P("lbkt", /\\int/), P("val"), P("rbkt"), P._, function (_) {
            throw new UnsupportedMathError("int");
        }).when(P._, "|", P("val"), "|", P._, function (_) {
            return apply1_f.bind(undefined, Math.abs, _.val);
        }).when(P._, /\.$/, P("val"), P("rbkt"), P._, function (_) {
            throw new UnsupportedMathError("hidden bracket");
        }).when(P._, P("lbkt"), P("val"), /^\./, P._, function (_) {
            throw new UnsupportedMathError("hidden bracket");
        }).when(P._, P("lbkt"), P("val"), P("rbkt"), P._, function (_) {
            return _.val;
        }).end;
        Visitor.prototype.visit_matrix = function (exp, elements) {
            throw new UnsupportedMathError("matrix");
        };

        var toFunction = function toFunction(math, options) {
            options = options || {};
            var oldyy = parser.yy;
            var errors = [];
            try {
                parser.yy = new Simplifier(options);
                var visitor = new Visitor(options);
                var exp = parser.parse(math);
                var func = exp.accept(visitor);
            } finally {
                parser.yy = oldyy;
            }
            return func;
        };
        toFunction.UnsupportedMathError = UnsupportedMathError;

        return {
            toFunction: toFunction
        };
    };

    // export module depending on what kind of environment we are in
    if (typeof exports === "object" && typeof module !== "undefined" && module.exports) {
        // commonjs (e.g. NodeJS)
        var smarttype = require("./smarttype");
        var simplifier = require("./simplifier");
        var P = require("P.js");
        var exp = factory(smarttype, simplifier, P);
        for (var key in exp) {
            exports[key] = exp[key];
        }
    } else if (typeof this.define === "function" && this.define.amd) {
        // AMD (e.g. requirejs)
        this.define("smarttype/func", ["smarttype", "smarttype/simplifier", "Pjs"], function (smarttype, x, Pjs) {
            return {
                toFunction: factory(smarttype, smarttype, Pjs).toFunction
            };
        });
    } else {
        // browser
        this.smarttype = this.smarttype || {};
        this.smarttype.toFunction = factory(this.smarttype, this.smarttype, this.Pjs).toFunction;
    }
}).call(this);
