/* eslint-env es6 */
/* global exports module require */
/* eslint new-cap: 0, no-multi-spaces: 0, key-spacing: 0 */
/**
 * Converts SmartType to LaTeX
 */
(function () {
    'use strict';

    var factory = function factory(smarttype, simplifier, P) {
        var parser = smarttype.parser;
        var Simplifier = simplifier.Simplifier;

        var Visitor = function Visitor(options) {
            this.options = options;
        };
        Visitor.prototype.visit_addition = function () {
            return Array.prototype.slice.call(arguments, 1).join('');
        };
        Visitor.prototype.visit_addition_sign = function (sign) {
            return sign[0] === '\\' ? sign + ' ' : sign;
        };
        Visitor.prototype.visit_multiplication = function () {
            return Array.prototype.slice.call(arguments, 1).join('');
        };
        Visitor.prototype.visit_multiplication_op = function (visible) {
            return visible ? '\\cdot ' : '';
        };
        var binop_map = {
            '==>': '\\Rightarrow ',
            '<=>': '\\Leftrightarrow ',
            '=^o': '\\circeq ',
            '.=': '\\doteq ',
            'in': '\\in ',
            '<-': '\\leftarrow ',
            '->': '\\rightarrow ',
            'le': '\\leq ',
            '<=': '\\leq ',
            '=<': '\\leq ',
            'ge': '\\geq ',
            '>=': '\\geq ',
            '=>': '\\geq ',
            '~=': '\\cong ',
            'approx': '\\approx ',
            'sub': '\\subset ',
            'sup': '\\supset ',
            'sube': '\\subseteq ',
            'supe': '\\supseteq ',
            'ne': '\\neq ',
            '/==>': '\\nRightarrow ',
            '/<=>': '\\nLeftrightarrow ',
            '/=': '\\neq ',
            '/<': '\\not<',
            '/>': '\\not>',
            '/in': '\\not\\in ',
            '/<-': '\\nleftarrow ',
            '/->': '\\nrightarrow ',
            '/le': '\\nleq ',
            '/<=': '\\nleq ',
            '/=<': '\\nleq ',
            '/ge': '\\ngeq ',
            '/>=': '\\ngeq ',
            '/=>': '\\ngeq ',
            '/~=': '\\ncong ',
            '/approx': '\\not\\approx ',
            '/sub': '\\not\\subset ',
            '/sup': '\\not\\supset ',
            '/sube': '\\not\\subseteq ',
            '/supe': '\\not\\supseteq ',
            'U': '\\cup ',
            'cap': '\\cap '
        };
        Visitor.prototype.visit_binop = new P.Function().when(P._, P('left'), 'root', P('right'), function (_) {
            return '\\sqrt[' + _.left + ']{' + _.right + '}';
        }).when(P._, P('left'), '_', P('right'), function (_) {
            return _.left + '_{' + _.right + '}';
        }).when(P._, P('left'), '^', P('right'), function (_) {
            return _.left + '^{' + _.right + '}';
        }).when(P._, P('left'), '/', P('right'), function (_) {
            return '\\frac{' + _.left + '}{' + _.right + '}';
        }).when(P._, P('left'), P('op'), P('right'), function (_) {
            return _.left + (_.op in binop_map ? binop_map[_.op] : _.op) + _.right;
        }).end;
        var nonstandard_func = {
            sech: true,
            csch: true,
            arcsech: true,
            arccsch: true,
            arccoth: true
        };
        Visitor.prototype.visit_funcname = function (name) {
            if (name in nonstandard_func) {
                return '\\mathrm{' + name + '}';
            }
            return '\\' + name + ' ';
        };
        Visitor.prototype.visit_bigop = new P.Function().when(P._, P('op'), null, null, function (_) {
            return '\\' + _.op + ' ';
        }).when(P._, P('op'), P('sub'), null, function (_) {
            return '\\' + _.op + '_{' + _.sub + '}';
        }).when(P._, P('op'), null, P('sup'), function (_) {
            return '\\' + _.op + '^{' + _.sup + '}';
        }).when(P._, P('op'), P('sub'), P('sup'), function (_) {
            return '\\' + _.op + '_{' + _.sub + '}' + '^{' + _.sup + '}';
        }).end;
        Visitor.prototype.visit_sign = function (exp, sign, value) {
            return sign + value;
        };
        Visitor.prototype.visit_func = new P.Function().when(P._, '\\floor ', P('value'), function (_) {
            return '\\left\\lfloor ' + _.value + '\\right\\rfloor ';
        }).when(P._, '\\ceil ', P('value'), function (_) {
            return '\\left\\lceil ' + _.value + '\\right\\rceil ';
        }).when(P._, '\\sqrt ', P('value'), function (_) {
            return '\\sqrt{' + _.value + '}';
        }).when(P._, P('name'), P('value'), function (_) {
            return _.name + _.value;
        }).end;
        Visitor.prototype.visit_postfunc = new P.Function().when(P._, '->', P('value'), function (_) {
            return '\\vec{' + _.value + '}';
        }).when(P._, '_', P('value'), function (_) {
            return '\\bar{' + _.value + '}';
        }).when(P._, '^', P('value'), function (_) {
            return '\\hat{' + _.value + '}';
        }).when(P._, '~', P('value'), function (_) {
            return '\\tilde{' + _.value + '}';
        }).when(P._, '.', P('value'), function (_) {
            return '\\dot{' + _.value + '}';
        }).when(P._, 'factorial', P('value'), function (_) {
            return _.value + '!';
        }).when(P._, 'prime', P.captureRE(/^(\{.*\^\{(?:\\prime)+)\}\}/, 'pre'), function (_) {
            return _.pre + '\\prime}}';
        }) // a hacky way to handle multiple primes
        .when(P._, 'prime', P('value'), function (_) {
            return '{' + _.value + '^{\\prime}}';
        }).when(P._, P('name'), P('value'), function (_) {
            return _.value + _.name;
        }).end;
        Visitor.prototype.visit_number = function (num) {
            return num.value;
        };
        Visitor.prototype.visit_variable = function (variable) {
            return variable.name;
        };
        var greek = /^(?:alpha|beta|[Gg]amma|[Dd]elta|(?:var)?epsilon|(?:z||(?:T|(?:var)?t)h)eta|iota|(?:var)?kappa|[Ll]ambda|[mn]u|[Xx]i|omicron|(?:P|(?:var)?p)i|(?:var)?rho|(?:S|(?:var)?s)igma|tau|[Uu]psilon|(?:(?:P|(?:var)?p)|c)hi|[Pp]si|[Oo]mega)$/;
        Visitor.prototype.visit_symbol = new P.Function().when({ name: null }, function (_) {
            return this.options.dispmissing ? '\\Box ' : '{}';
        }).when({ name: P('name', greek) }, function (_) {
            return '\\' + _.name + ' ';
        }).when({ name: 'd.' }, function (_) {
            return '\\mathrm{d}';
        }).when({ name: P.captureRE(/^d(.)$/, 'x') }, function (_) {
            return '\\mathrm{d}' + _.x;
        }).when({ name: 'inf' }, function (_) {
            return '\\infty ';
        }).when({ name: P('name', /^Z[-+]$/) }, function (_) {
            return '\\mathbb{Z}^' + _.name[1];
        }).when({ name: '{}' }, function (_) {
            return '\\emptyset ';
        }).when({ name: '...' }, function (_) {
            return '\\dots ';
        }).when({ name: 'undefined' }, function (_) {
            return '\\textrm{undefined}';
        }).when({ name: 'doesnotexist' }, function (_) {
            return '\\textrm{doesnotexist}';
        }).when({ name: P('name', /^(?:CC|QQ|NN|RR|ZZ)$/) }, function (_) {
            return '\\mathbb{' + _.name[0] + '}';
        }).when({ name: P('name') }, function (_) {
            return _.name;
        }).end;
        Visitor.prototype.visit_bracket = new P.Function()
        // unprocessed integral
        .when(P._, 'int', P('val'), P('rbkt'), P._, function (_) {
            return '\\int ' + _.val + '\\;' + this.visit_symbol({ name: _.rbkt });
        })
        // already processed integral
        .when(P._, P('lbkt', /\\int/), P('val'), P('rbkt'), P._, function (_) {
            return _.lbkt + _.val + '\\;' + this.visit_symbol({ name: _.rbkt });
        }).when(P._, /\.$/, P('val'), /^\./, P._, function (_) {
            return _.val;
        }).when(P._, /\.$/, P('val'), '}', P._, function (_) {
            return '\\left.' + _.val + '\\right\\}';
        }).when(P._, /\.$/, P('val'), P('rbkt'), P._, function (_) {
            return '\\left.' + _.val + '\\right' + _.rbkt;
        }).when(P._, '{', P('val'), /^\./, P._, function (_) {
            return '\\left\\{' + _.val + '\\right.';
        }).when(P._, P('lbkt'), P('val'), /^\./, P._, function (_) {
            return '\\left' + _.lbkt + _.val + '\\right.';
        }).when(P._, '{', P('val'), '}', P._, function (_) {
            return '\\left\\{' + _.val + '\\right\\}';
        }).when(P._, P('lbkt'), P('val'), '}', P._, function (_) {
            return '\\left' + _.lbkt + _.val + '\\right\\}';
        }).when(P._, '{', P('val'), P('rbkt'), P._, function (_) {
            return '\\left\\{' + _.val + '\\right' + _.rbkt;
        }).when(P._, P('lbkt'), P('val'), P('rbkt'), P._, function (_) {
            return '\\left' + _.lbkt + _.val + '\\right' + _.rbkt;
        }).end;
        Visitor.prototype.visit_matrix = function (exp, elements) {
            var result = '\\begin{matrix} ' + elements[0].join(' & ');
            for (var i = 1; i < elements.length; i++) {
                var row = elements[i];
                result += ' \\\\ ' + row.join(' & ');
            }
            result += ' \\end{matrix}';
            return result;
        };

        var toLaTeX = function toLaTeX(math, options) {
            options = options || {};
            var oldyy = parser.yy;
            var errors = [];
            try {
                parser.yy = new Simplifier(options);
                var visitor = new Visitor(options);
                var exp = parser.parse(math);
                var latex = exp.accept(visitor);
            } finally {
                parser.yy = oldyy;
            }
            return { latex: latex, errors: errors };
        };

        return {
            toLaTeX: toLaTeX
        };
    };

    // export module depending on what kind of environment we are in
    if (typeof exports === 'object' && typeof module !== 'undefined' && module.exports) {
        // commonjs (e.g. NodeJS)
        var smarttype = require('./smarttype');
        var simplifier = require('./simplifier');
        var P = require('P.js');
        var exp = factory(smarttype, simplifier, P);
        for (var key in exp) {
            exports[key] = exp[key];
        }
    } else if (typeof this.define === 'function' && this.define.amd) {
        // AMD (e.g. requirejs)
        this.define('smarttype/latex', ['smarttype', 'smarttype/simplifier', 'Pjs'], function (smarttype, x, Pjs) {
            return {
                toLaTeX: factory(smarttype, smarttype, Pjs).toLaTeX
            };
        });
    } else {
        // browser
        this.smarttype = this.smarttype || {};
        this.smarttype.toLaTeX = factory(this.smarttype, this.smarttype, this.Pjs).toLaTeX;
    }
}).call(this);
