    // Trim off spaces
    parser.parse = function (oldParse)
    {
        "use strict";
        return function (input) {
            return oldParse.call(this, input.trim());
        }
    }(parser.parse);

    // Do some cheating in the lexer by keeping some state, in order to resolve
    // some ambiguities in the parser.  Monkey patch lexer's setInput method
    // (to initialize state), and lex method (to match brackets and integrals)
    parser.lexer.setInput = function (oldSetInput)
    {
        "use strict";
        return function () {
            oldSetInput.apply(this, Array.prototype.slice.call(arguments));
            this.smarttype = {
                brackets: [],
                prev: undefined,
                next: [],
                matched: ""
            };
        };
    }(parser.lexer.setInput);

    parser.lexer.lex = function (oldLex)
    {
        "use strict";
        return function () {
            var state = this.smarttype;
            var token;
            if (state.next.length)
            {
                var next = state.next.shift();
                for (var name in next[1])
                {
                    this[name] = next[1][name];
                }
                token = next[0];
            }
            else
            {
                token = oldLex.call(this);
            }
            if (typeof token !== 'number')
            {
                token = parser.symbols_[token] || token;
            }
            var tokenname = parser.terminals_[token] || token;
            switch (tokenname) {
            case 'SPCINT':
                state.next.push(['INT', {match: 'int', yytext: 'int'}]);
                tokenname = 'SPC';
                token = parser.symbols_['SPC'];
                this.match = this.yytext = ' ';
                break;
            case 'SPCSYMBOL':
                var trimmed = this.match.trim();
                state.next.push(['SYMBOL', {match: trimmed, yytext: trimmed}]);
                tokenname = 'SPC';
                token = parser.symbols_['SPC'];
                this.match = this.yytext = ' ';
                break;
            case 'SPCUNDEFINED':
                var trimmed = this.match.trim();
                state.next.push(['UNDEFINED', {match: trimmed, yytext: trimmed}]);
                tokenname = 'SPC';
                token = parser.symbols_['SPC'];
                this.match = this.yytext = ' ';
                break;
            case 'SPCDOESNOTEXIST':
                var trimmed = this.match.trim();
                state.next.push(['DOESNOTEXIST', {match: trimmed, yytext: trimmed}]);
                tokenname = 'SPC';
                token = parser.symbols_['SPC'];
                this.match = this.yytext = ' ';
                break;
            case 'LBRACKET':
            case 'LPIPE':
            case 'INT':
                // track brackets and integrals
                state.brackets.push(tokenname);
                break;
            case 'RBRACKET':
                // match closing brackets with opening brackets
                if (state.brackets.length)
                {
                    while (state.brackets[state.brackets.length - 1] === 'INT')
                    {   // a closing bracket will also close an integral
                        state.brackets.pop();
                    }
                    state.brackets.pop();
                }
                break;
            case '|':
                if (state.brackets.length
                    && state.brackets[state.brackets.length - 1] === 'LPIPE')
                {   // matching an opening pipe
                    state.brackets.pop();
                    tokenname = 'RPIPE';
                    token = parser.symbols_['RPIPE'];
                }
                else
                {   // no opening pipe to match, so assume we are an opening pipe
                    state.brackets.push('LPIPE');
                    tokenname = 'LPIPE';
                    token = parser.symbols_['LPIPE'];
                }
                break;
            case 'DX':
                // dx can either be used to close an integral, or as part of a
                // derivative.  If we are in an integral, we grab the first dx
                // that we can (that is, the first dx that is in the same
                // bracket level, and that does not immediately follow a '/' or
                // another 'dx') to close the integral.  We ignore dx's that
                // immediately follow a '/' to more easily allow for
                // derivatives inside of integrals.
                //
                // One tricky thing is that in order to get around a parser
                // issue, dx swallows leading spaces.  However, spaces can be
                // significant in parsing, so if we are not expecting to close
                // an integral, we split it into two tokens
                if (this.match[0] === ' ')
                {
                    if (state.brackets.length
                        && state.brackets[state.brackets.length - 1] === 'INT'
                        && state.prev !== '/')
                    {
                        state.brackets.pop();
                        tokenname = 'IDX';
                        token = parser.symbols_['IDX'];
                        this.match = this.yytext = this.match.slice(-2);
                    }
                    else
                    {
                        var match = this.match.slice(-2);
                        state.next.push(['DX', {match: match, yytext: match}]);
                        tokenname = 'SPC';
                        token = parser.symbols_['SPC'];
                    }
                }
                else if (state.brackets.length
                            && state.brackets[state.brackets.length - 1] === 'INT'
                            && state.prev !== '/' && state.prev !== 'DX')
                {   // if we are in an integral (and not right after a slash),
                    // then close the integral
                    state.brackets.pop();
                    tokenname = 'IDX';
                    token = parser.symbols_['IDX'];
                }
                break;
            }
            state.prev = tokenname;
            return token;
        };
    }(parser.lexer.lex);

    var cloneobj = function (obj)
    {
        var result = {};
        for (var k in obj)
        {
            result[k] = obj[k];
        }
        return result;
    };

    // Cheat and monkey patch some more in order to fix some parse errors.
    // - match parentheses
    // - add in missing values
    parser.lexer.setInput = function (oldSetInput)
    {
        "use strict";
        return function () {
            oldSetInput.apply(this, Array.prototype.slice.call(arguments));
            var tokens = this.smarttype.tokens = [];
        };
    }(parser.lexer.setInput);

    // tokens what are considered values when seen on the given side
    var values = {
        left: {
            NUMBER: true,
            VARIABLE: true,
            MVARIABLE: true,
            SYMBOL: true,
            UNDEFINED: true,
            DOESNOTEXIST: true,
            DX: true,
            MISSINGVALUE: true,
            LBRACKET: true,
            LPIPE: true,
            INT: true,
            FUNCSUB: true,
            FUNCSUP: true,
            FUNC: true,
            BIGOP: true,
            PM: true,
            MP: true,
            "+": true,
            "-": true,
            LIM: true
        },
        right: {
            NUMBER: true,
            VARIABLE: true,
            MVARIABLE: true,
            SYMBOL: true,
            UNDEFINED: true,
            DOESNOTEXIST: true,
            DX: true,
            MISSINGVALUE: true,
            RBRACKET: true,
            RPIPE: true,
            IDX: true,
            PRIME: true,
            SUPDECORATION: true,
            "!": true,
            "?": true
        }
    };

    // whether the given token expects a value (or something else) on the given side
    var token_expects_value = {
        left: {
            SUPDECORATION: true,
            ",": true,
            RELATION: true,
            SETOP: true,
            BACKSLASH: true,
            "&": true,
            "/": true,
            "*": true,
            ROOT: true,
            "^": {FUNCSUP: true, INT: true, BIGOP: true},
            "_": {FUNCSUB: true, INT: true, BIGOP: true, LIM: true},
            ",": true,
            "!": true,
            PRIME: true,
            RBRACKET: true,
            RPIPE: true,
            IDX: true,
            EOF: true
        },
        right: {
            ",": true,
            RELATION: true,
            SETOP: true,
            BACKSLASH: true,
            "&": true,
            "/": true,
            "*": true,
            ROOT: true,
            "^": true,
            "_": true,
            ",": true,
            LBRACKET: true,
            LPIPE: true
        }
    };

    var bracketmatch = {
        left: {
            ")": "(",
            "]": "[",
            "}": "{",
            "|": "|",
            ".)": "(",
            ".]": "[",
            ".}": "{",
            ".|": "|"
        },
        right: {
            "(": ")",
            "{": "}",
            "[": "]",
            "|": "|",
            "(.": ")",
            "{.": "}",
            "[.": "]",
            "|.": "|",
            "int": ""
        }
    };

    var tokenmatch = {
        left: {
            RPIPE: parser.symbols_.LPIPE,
            RBRACKET: parser.symbols_.LBRACKET
        },
        right: {
            LPIPE: parser.symbols_.RPIPE,
            LBRACKET: parser.symbols_.RBRACKET,
            INT: parser.symbols_.IDX
        }
    }

    // patch the tokens array to fix patch errors
    var fix_tokens = function (tokens)
    {
        "use strict";
        var to_insert = []; // extra tokens to insert into the tokens array
        var prev; // previous token found
        var expect_value = true; // whether we expect the next token to a value
        var bracket_depth = 0; // the current depth of nested brackets
        var bracket_stack = []; // the pending open brackets
        var eofidx; // the index of the EOF token
        for (var i = 0; i < tokens.length; i++)
        {
            var token = tokens[i];
            var tokenname = parser.terminals_[token.token] || token.token;
            if (tokenname === "SPC")
            {   // ignore spaces
                continue;
            }
            else if (tokenname === "EOF")
            {
                eofidx = i;
            }
            if (expect_value && !(tokenname in values.left))
            {   // value expected, but not found: add a "missing value" token
                to_insert.push({idx: i,
                                token: { token: parser.symbols_.MISSINGVALUE,
                                            yytext: "",
                                            yyloc: { first_column: token.yyloc.first_column,
                                                    first_line: token.yyloc.first_line,
                                                    last_line: token.yyloc.first_column,
                                                    last_column: token.yyloc.first_line }}});
                prev = "MISSINGVALUE";
            }
            if (tokenname in token_expects_value.left
                && !(prev in values["right"]
                        || (typeof token_expects_value.left[tokenname] === "object"
                            && prev in token_expects_value.left[tokenname])))
            {   // see if anything is expected on the left: if not found, add a "missing value" token
                to_insert.push({idx: i,
                                token: { token: parser.symbols_.MISSINGVALUE,
                                            yytext: "",
                                            yyloc: { first_column: token.yyloc.first_column,
                                                    first_line: token.yyloc.first_line,
                                                    last_line: token.yyloc.first_column,
                                                    last_column: token.yyloc.first_line }}});
            }
            switch (tokenname) {
            case "?":
                if (prev !== "RBRACKET" && prev !== "RPIPE")
                {
                    // FIXME:
                }
                break;
            case "RBRACKET":
            case "RPIPE":
            case "IDX":
                // found right bracket: try to match left bracket
                bracket_depth--;
                if (bracket_stack.length)
                {
                    var lbracket = bracket_stack.pop();
                    while (lbracket && lbracket.token === "INT" && tokenname !== "IDX")
                    {   // if the left bracket is an integral, and we're a normal bracket, add the missing dx
                        to_insert.push({idx: i,
                                        token: { token: parser.symbols_.PREEND,
                                                    yytext: "",
                                                    yyloc: { first_column: token.yyloc.first_column,
                                                            first_line: token.yyloc.first_line,
                                                            last_line: token.yyloc.first_column,
                                                            last_column: token.yyloc.first_line }}},
                                        {idx: i,
                                        token: { token: parser.symbols_.IDX,
                                                    yytext: "",
                                                    yyloc: { first_column: token.yyloc.first_column,
                                                            first_line: token.yyloc.first_line,
                                                            last_line: token.yyloc.first_column,
                                                            last_column: token.yyloc.first_line }}});
                        lbracket = bracket_stack.pop();
                        bracket_depth--;
                    }
                }
                if (bracket_depth < 0)
                {   // too many closing brackets -- add an opening bracket at the beginning
                    to_insert.unshift({idx: 0,
                                        token: {token: tokenmatch.left[tokenname],
                                                yytext: bracketmatch.left[token.yytext],
                                                yyloc: {first_column: 0,
                                                        first_line: 0,
                                                        last_line: 0,
                                                        last_column: 0}}});
                    bracket_depth++;
                    // FIXME: signal error
                }
                to_insert.push({idx: i,
                                token: { token: parser.symbols_.PREEND,
                                            yytext: "",
                                            yyloc: { first_column: token.yyloc.first_column,
                                                    first_line: token.yyloc.first_line,
                                                    last_line: token.yyloc.first_column,
                                                    last_column: token.yyloc.first_line }}})
                break;
            case "LBRACKET":
            case "LPIPE":
            case "INT":
                bracket_depth++;
                bracket_stack.push({token: tokenname, yytext: token.yytext});
                break;
            }
            expect_value = tokenname in token_expects_value["right"];
            prev = tokenname;
        }
        while (bracket_stack.length)
        {   // unmatched opening brackets: add matching closing brackets at end
            var token = bracket_stack.pop();
            to_insert.push({idx: eofidx,
                            token: { token: parser.symbols_.PREEND,
                                        yytext: "",
                                        yyloc: { first_column: 0,
                                                first_line: 0,
                                                last_line: 0,
                                                last_column: 0 }}},
                            {idx: eofidx,
                            token: {token: tokenmatch.right[token.token],
                                    yytext: bracketmatch.right[token.yytext],
                                    yyloc: {first_column: 0,
                                            first_line: 0,
                                            last_line: 0,
                                            last_column: 0}}});
            // FIXME: signal error
        }
        to_insert.push({idx: eofidx,
                        token: { token: parser.symbols_.PREEND,
                                    yytext: "",
                                    yyloc: { first_column: 0,
                                            first_line: 0,
                                            last_line: 0,
                                            last_column: 0 }}})

        // insert the extra tokens into the tokens array
        // NOTE: start from the last token and go to the first, so that we don't
        // mess up the array indices
        while (to_insert.length)
        {
            var ins = to_insert.pop();
            tokens.splice(ins.idx, 0, ins.token);
        }
    };

    parser.lexer.lex = function (oldLex)
    {
        "use strict";
        return function () {
            var tokens = this.smarttype.tokens;
            if (tokens.length === 0)
            {
                var token, yyloc;
                var newlex = this.lex;
                this.lex = oldLex;
                while ((token = this.lex(this)) !== this.EOF)
                {
                    tokens.push({token: token, yytext: this.yytext, yyloc: cloneobj(this.yylloc)});
                }
                tokens.push({token: token, yytext: this.yytext, yyloc: cloneobj(this.yyloc)});
                this.lex = newlex;
                fix_tokens(tokens);
            }
            var r = tokens[0];
            if (tokens.length > 1)
            {
                tokens.shift();
            }
            this.smarttype.matched += r.yytext;
            this.yytext = this.match = r.yytext;
            this.yyloc = this.yylloc = r.yyloc;
            return r.token;
        }
    }(parser.lexer.lex);

    parser.lexer.pastInput = function (oldPastInput)
    {
        "use strict";
        return function () {
            var matched = this.smarttype.matched;
            var past = matched.substr(0, matched.length - this.match.length);
            return (past.length > 20 ? '...':'') + past.substr(-20).replace(/\n/g, "");
        }
    }(parser.lexer.pastInput)

    // export module depending on what kind of environment we are in
    if (typeof exports === 'object' && typeof module !== 'undefined' && module.exports) {
        // commonjs (e.g. NodeJS)
        exports.parser = parser;
    } else if (typeof this.define === 'function' && this.define.amd) {
        // AMD (e.g. requirejs)
        this.define('smarttype', function () {
            return {
                parser: parser
            };
        });
    } else {
        // browser
        this.smarttype = this.smarttype || {};
        this.smarttype.parser = parser;
    }

