/* simplifier es6 */
/* global exports module require */
/* eslint new-cap: 0, camelcase: 0, quotes: 0 */
/**
 * Simplifies a SmartType parse tree for to Javascript Object used by latex.es6.js.
 */
(function () {
    "use strict";

    var factory = function (P)
    {
        var init = function (type, data) {
            this.type = type;
            for (var key in data)
            {
                this[key] = data[key];
            }
        };

        var classes = {
            Addition: function (data) {
                init.call(this, 'addition', data);
            },
            Multiplication: function (data) {
                init.call(this, 'multiplication', data);
            },
            Binop: function (data) {
                init.call(this, 'binop', data);
            },
            Bigop: function (data) {
                init.call(this, 'bigop', data);
            },
            Sign: function (data) {
                init.call(this, 'sign', data);
            },
            Func: function (data) {
                init.call(this, 'func', data);
            },
            Postfunc: function (data) {
                init.call(this, 'postfunc', data);
            },
            Number: function (data) {
                init.call(this, 'number', data);
            },
            Input: function (data) {
                init.call(this, 'input', data);
            },
            Latex: function (data) {
                init.call(this, 'latex', data);
            },
            Symbol: function (data) {
                init.call(this, 'symbol', data);
            },
            Variable: function (data) {
                init.call(this, 'variable', data);
            },
            Bracket: function (data) {
                init.call(this, 'bracket', data);
            },
            Matrix: function (data) {
                init.call(this, 'matrix', data);
            }
        };

        // accept methods for the visitor pattern
        // https://en.wikipedia.org/wiki/Visitor_pattern
        classes.Addition.prototype.accept = function (visitor)
        {
            var args = [this];
            args.push(this.operands[0] === null ? null : this.operands[0].accept(visitor));
            for (var i = 1; i < this.operands.length; i++)
            {
                args.push(visitor.visit_addition_sign(this.signs[i-1]));
                args.push(this.operands[i] === null ? null : this.operands[i].accept(visitor));
            }
            return visitor.visit_addition.apply(visitor, args);
        };
        classes.Multiplication.prototype.accept = function (visitor)
        {
            var args = [this];
            args.push(this.operands[0] === null ? null : this.operands[0].accept(visitor));
            for (var i = 1; i < this.operands.length; i++)
            {
                args.push(visitor.visit_multiplication_op(this.visible[i-1]));
                args.push(this.operands[i] === null ? null : this.operands[i].accept(visitor));
            }
            return visitor.visit_multiplication.apply(visitor, args);
        };
        classes.Binop.prototype.accept = function (visitor)
        {
            return visitor.visit_binop(this, typeof this.left === 'string' ? visitor.visit_funcname(this.left) : this.left.accept(visitor), this.op, this.right.accept(visitor));
        };
        classes.Bigop.prototype.accept = function (visitor)
        {
            return visitor.visit_bigop(this, this.op, this.sub === null ? null : this.sub.accept(visitor), this.sup === null ? null : this.sup.accept(visitor));
        };
        classes.Sign.prototype.accept = function (visitor)
        {
            return visitor.visit_sign(this, this.op, this.value.accept(visitor));
        };
        classes.Func.prototype.accept = function (visitor)
        {
            return visitor.visit_func(this, typeof this.name === 'string' ? visitor.visit_funcname(this.name) : this.name.accept(visitor), this.value.accept(visitor));
        };
        classes.Postfunc.prototype.accept = function (visitor)
        {
            return visitor.visit_postfunc(this, this.name, this.value.accept(visitor));
        };
        classes.Number.prototype.accept = function (visitor)
        {
            return visitor.visit_number(this);
        };
        classes.Input.prototype.accept = function (visitor)
        {
            return visitor.visit_input(this );
        };
        classes.Latex.prototype.accept = function (visitor)
        {
            return visitor.visit_latex(this );
        };
        classes.Variable.prototype.accept = function (visitor)
        {
            return visitor.visit_variable(this);
        };
        classes.Symbol.prototype.accept = function (visitor)
        {
            return visitor.visit_symbol(this);
        };
        classes.Bracket.prototype.accept = function (visitor)
        {
            return visitor.visit_bracket(this, typeof this.lbracket === 'string' ? this.lbracket : this.lbracket.accept(visitor), this.value.accept(visitor), this.rbracket, this.hidden);
        };
        classes.Matrix.prototype.accept = function (visitor)
        {
            var matrixargs = [];
            for (var i = 0; i < this.elements.length; i++)
            {
                var row = [];
                matrixargs.push(row);
                for (var j = 0; j < this.elements[i].length; j++)
                {
                    row.push(this.elements[i][j] === null ? null : this.elements[i][j].accept(visitor));
                }
            }
            return visitor.visit_matrix(this, matrixargs);
        };

        // functions to build and simplify expression tree
        var atom = /^(symbol|number|variable)$/;
        var sub_op = /^(?:-|\\pm|\\mp)$/;
        var mult_op = /^(?:\*|\(\*\))$/;
        var lparen = /^[\[\(\{]$/, rparen = /^[\]\)\}]$/;
        // atoms don't need parentheses around them in some contexts
        var atomic_operand = new P.Function()
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: atom})}, _=> _.op)
            .when(P('op'), _ => _.op)
            .end;
        // addition operands: a bracketed addition, or anything with higher precedence, should be lifted out of the brackets
        var addition_operands = new P.Function()
            .when(P('op', {type: 'addition'}), _ => _.op)
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'addition'})},
                  _ => _.op)
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'multiplication'})},
                  _ => ({operands: [_.op], signs: []}))
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'binop', op: /^([\/\^_]|root)$/})},
                  _ => ({operands: [_.op], signs: []}))
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'func'})},
                  _ => ({operands: [_.op], signs: []}))
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: atom})},
                  _ => ({operands: [_.op], signs: []}))
            .when(P('op'), _ => ({operands: [_.op], signs: []}))
            .end;
        // subtraction operand (rhs): similar to addition operands, but don't lift bracketed addition
        var subtraction_operands = new P.Function()
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'multiplication'})},
                  _ => [_.op])
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'binop', op: /^([\/\^_]|root)$/})},
                  _ => [_.op])
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'func'})},
                  _ => [_.op])
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: atom})},
                  _ => [_.op])
            .when(P('op'), _ => [_.op])
            .end;
        // multiplication operands: a bracketed multiplication, or anything with higher precedence, should be lifted out of the brackets
        var multiplication_operands = new P.Function()
            .when(P('op', {type: 'multiplication'}), _ => _.op)
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'multiplication'})},
                  _ => _.op )
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'binop', op: /^([\/^_]|root)$/})},
                  _ => ({operands: [_.op], visible: []}))
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: 'func'})},
                  _ => ({operands: [_.op], visible: []}))
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op', {type: atom})},
                  _ => ({operands: [_.op], visible: []}))
            .when(P('op'), _ => ({operands: [_.op], visible: []}))
            .end;
        // strip off parentheses, if present
        var strip_parens = new P.Function()
        // don't strip parens off of lists, columns, rows, setop
            .when(P('op', {type: 'bracket', value: {type: 'binop', op: /^([,&\\U]|cap)$/}}), _ => _.op)
            .when({type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('op')}, _ => _.op)
            .when(P('op'), _ => _.op)
            .end;

        function mult_visible(left, right)
        {
            var last = left[left.length - 1];
            var first = right[0];
            return ((last.type === 'symbol' && last.name === null) || last.type === 'number')
                && ((first.type === 'symbol' && first.name === null) || (first.type === 'number') || (first.type === 'binop' && first.op === '^' && first.left.type === 'number'));
        }
        // FIXME: auto-paren vs manual paren
        var binop = new P.Function()
            .when(P('e1'), '+', P('e2'), function (_) {
                var operands1 = addition_operands(_.e1), operands2 = addition_operands(_.e2);
                return new classes.Addition({operands: operands1.operands.concat(operands2.operands), signs: operands1.signs.concat('+', operands2.signs)});
            })
            .when(P('e1'), P('op', sub_op), P('e2'), function (_) {
                var operands1 = addition_operands(_.e1), operands2 = subtraction_operands(_.e2);
                return new classes.Addition({operands: operands1.operands.concat(operands2), signs: operands1.signs.concat(_.op)});
            })
            .when(P('e1'), '(*)', P('e2', {type: 'multiplication'}), function (_) {
                var operands1 = multiplication_operands(_.e1);
                return new classes.Multiplication({operands: operands1.concat(_.e2.operands),
                                                   visible: operands1.visible.concat(mult_visible(operands1, _.e2.operands),
                                                                                     _.e2.visible)});
            })
            .when(P('e1'), '(*)', P('e2'), function (_) {
                // FIXME: distinguish between space and no-space implicit multiplication

                // implicit multiplication should retain parentheses on the
                // right hand side, so that "f(x)" (which is parsed as an
                // implicit multiplication) displays properly
                var operands1 = multiplication_operands(_.e1);
                return new classes.Multiplication({operands: operands1.operands.concat([_.e2]),
                                                   visible: operands1.visible.concat(mult_visible(operands1.operands,
                                                                                                  [_.e2]))});
            })
            .when(P('e1'), P('op', mult_op), P('e2'), function (_) {
                var operands1 = multiplication_operands(_.e1), operands2 = multiplication_operands(_.e2);
                return new classes.Multiplication({operands: operands1.operands.concat(operands2.operands),
                                                   visible: operands1.visible.concat(mult_visible(operands1.operands, operands2.operands),
                                                                                     operands2.visible)});
            })
            .when(P('e1'), P('op', /^(\/|root)$/), P('e2'), function (_) {
                return new classes.Binop({op: _.op, left: strip_parens(_.e1), right: strip_parens(_.e2)});
            })
            .when(P('e1'), P('op', /^[\^_]$/), P('e2'), function (_) {
                return new classes.Binop({op: _.op, left: _.e1, right: strip_parens(_.e2)});
            })
            .when(P('matrix', {type: 'matrix'}), '&', P('e2'), function (_) {
                var M = _.matrix;
                var lastrow = M.elements[M.elements.length - 1];
                lastrow.push(_.e2);
                if (lastrow.length > M.width)
                {
                    M.width = lastrow.length;
                }
                return M;
            })
            .when(P('e1'), '&', P('e2'),
                  _ => new classes.Matrix({elements: [[_.e1, _.e2]], width: 2, height: 1}))
            .when(P('matrix', {type: 'matrix'}), '\\', P('e2', {type: 'matrix'}), function (_) {
                var M = _.matrix;
                M.elements.push(_.e2.elements[0]);
                M.height = M.elements.length;
                if (_.e2.elements[0].length > M.width)
                {
                    M.width = _.e2.elements[0].length;
                }
                return M;
            })
            .when(P('matrix', {type: 'matrix'}), '\\', P('e2'), function (_) {
                var M = _.matrix;
                M.elements.push([_.e2]);
                M.height = M.elements.length;
                return M;
            })
            .when(P('e1'), '\\', P('e2'),
                  _ => new classes.Matrix({elements: [[_.e1], [_.e2]], width: 1, height: 2}))
            .when(P('e1'), P('op'), P('e2'),
                  _ => new classes.Binop({op: _.op, left: _.e1, right: _.e2}))
            .end;
        var bigop = function (op, sub, sup)
        {
            return new classes.Bigop({op: op, sub: strip_parens(sub), sup: strip_parens(sup)});
        };
        var sign = new P.Function()//(-) ( [sum_{}^{} blah] )
            .when(P('op'), {type: 'bracket', lbracket: lparen, rbracket: rparen, value: P('e', {type: 'binop', op: '/'})},
                  _ => new classes.Sign({op: _.op, value: _.e}))
            .when(P('op'), P('e', {type: 'bracket', lbracket: lparen, rbracket: rparen}),
                  _ => new classes.Sign({op: _.op, value: _.e}))
            .when(P('op'), P('e', {type: atom}),
                  _ => new classes.Sign({op: _.op, value: _.e}))
            .when(P('op'), P('e'),
                  _ => new classes.Sign({op: _.op, value: new classes.Bracket({lbracket: '(', rbracket: ')', value: _.e, hidden: false})}))
            .end;
        var bigop_re = /^(int|lim|prod|sum)$/;
        // FIXME: auto-paren vs manual paren
        var func = new P.Function()
        // bigops should remain unchanged, except for bracketed atoms
            .when(P('name', bigop_re), P('e'), _ => new classes.Func({name: _.name, value: atomic_operand(_.e)}))
            .when(P('name', {type: 'bigop'}), P('e'), _ => new classes.Func({name: _.name, value: atomic_operand(_.e)}))
            .when(P('name', {type: 'binop', left: 'lim', op: '_'}), P('e'), _ => new classes.Func({name: _.name, value: atomic_operand(_.e)}))
        // square roots should be unbracketed
            .when('sqrt', P('e'), _ => new classes.Func({name: 'sqrt', value: strip_parens(_.e)}))
        // absolute values should be unbracketed, and put in pipes
            .when('abs', P('e'), _ => new classes.Bracket({lbracket: '|', rbracket: '|', value: strip_parens(_.e), hidden: false}))
        // normal functions:
        /* FIXME: in non-auto-paren mode, for normal functions, make sure e is
         * wrapped in a single set of parentheses, unless it is already
         * enclosed in brackets.
         */
        // function of an absolute value: leave as-is
            .when(P('name'), P('e', {type: 'bracket', lbracket: '|', rbracket: '|', hidden: false}),
                  _ => new classes.Func({name: _.name, value: _.e}))
        // function of curly braces: leave as-is
            .when(P('name'), P('e', {type: 'bracket', lbracket: '{', rbracket: '}', hidden: false}),
                  _ => new classes.Func({name: _.name, value: _.e}))
        // function of any other bracketed expression (not necessarily matched): strip brackets and wrap in parentheses
            .when(P('name'), {type: 'bracket', lbracket: /^[\(\[\{\|]$/, rbracket: /^[\)\]\}\|]$/, value: P('e')},
                  _ => new classes.Func({name: _.name, value: new classes.Bracket({lbracket: '(', rbracket: ')', value: _.e, hidden: false})}))
        // function of any other expression: wrap in parentheses
            .when(P('name'), P('e'),
                  _ => new classes.Func({name: _.name, value: new classes.Bracket({lbracket: '(', rbracket: ')', value: _.e, hidden: false})}))
            .end;
        var postfunc = function (name, e)
        {
            return new classes.Postfunc({name: name, value: e});
        };
        var number = function (value)
        {
            return new classes.Number({value: value});
        };
        var input = (value, index )=>{
            return new classes.Input({name : value});
        };
        var latex = (value, index)=>{
            return new classes.Latex({ latex : value.slice(1,-1)});
        };
        var variable = function (name)
        {
            return new classes.Variable({name: name});
        };
        var symbol = function (name)
        {
            return new classes.Symbol({name: name});
        };
        var bracket = new P.Function()
        // strip nested parentheses
            .when(lparen, {type: 'bracket',
                           lbracket: P('lbracket', lparen),
                           value: P('value'),
                           rbracket: P('rbracket', rparen),
                           hidden: P('hidden')}, rparen, P._,
                  _ => new classes.Bracket({lbracket: _.lbracket, value: _.value === null ? symbol(null) : _.value, rbracket: _.rbracket, hidden: !!_.hidden})
                  )
            .when(P('lbracket'), {type: 'bracket',
                                  lbracket: lparen,
                                  value: P('value'),
                                  rbracket: rparen}, P('rbracket'), P('hidden'),
                  _ => new classes.Bracket({lbracket: _.lbracket, value: _.value === null ? symbol(null) : _.value, rbracket: _.rbracket, hidden: !!_.hidden})
                 )
            .when(P('lbracket'), P('value'), P('rbracket'), P('hidden'),
                  _ => new classes.Bracket({lbracket: _.lbracket, value: _.value === null ? symbol(null) : _.value, rbracket: _.rbracket, hidden: !!_.hidden})
                 )
            .end;

        var Simplifier = function (options) {
            this.options = options;
            this.binop = binop;
            this.bigop = bigop;
            this.sign = sign;
            this.func = func;
            this.postfunc = postfunc;
            this.number = number;
            this.input = input;
            this.latex = latex;
            this.variable = variable;
            this.symbol = symbol;
            this.bracket = bracket;
            return this;
        };
        for (var cls in classes)
        {
            Simplifier[cls] = classes[cls];
        }

        return {
            Simplifier: Simplifier
        };
    };

    // export module depending on what kind of environment we are in
    if (typeof exports === 'object' && typeof module !== 'undefined' && module.exports) {
        // commonjs (e.g. NodeJS)
        var P = require('P.js');
        exports.Simplifier = factory(P).Simplifier;
    } else if (typeof this.define === 'function' && this.define.amd) {
            // AMD (e.g. requirejs)
            this.define('smarttype/simplifier', ['Pjs'], function (Pjs) {
                return {
                    Simplifier: factory(Pjs).Simplifier
                };
            });
    } else {
        // browser
        this.smarttype = this.smarttype || {};
        this.smarttype.Simplifier = factory(this.Pjs).Simplifier;
    }
}.call(this));
