/* -*- bison -*- */
/* description: parses SmartType */

/* lexical grammar */
%lex

%%
/* NOTE: we ignore spaces in the tokenizer where they should not be relevant.
 * e.g. before and after binary operators, after function names
 * i.e. wherever an implicit multiplication would not make sense */
"`"([^`]|"\\`")+"`"             return 'LATEX';
[0-9]+("."[0-9]+)?              return 'NUMBER';
"."[0-9]+                       return 'NUMBER';
"__("([A-Za-z.]*)")"            return 'INPUT';
(arc)?(sin|cos|tan|sec|cot|csc)h?\s* return 'FUNCSUP';
("log")\s*                      return 'FUNCSUB';
("ln"|"ceil"|"floor"|"sqrt"|"abs"|"exp"|"min"|"max")\s* return 'FUNC';
\s*("root")\s*                  return 'ROOT';
\s*"'"                          return 'PRIME';
("sum")\s*                      return 'BIGOP';
("prod")\s*                     return 'BIGOP';
("int")\s*                      return 'INT';
("lim")\s*                      return 'LIM';
("inf")                         return 'SYMBOL';
\s+("int")\s*                   return 'SPCINT'; // otherwise " int" is parsed as ("in", "t")
\s+("inf"|"delta"|"Upsilon")    return 'SPCSYMBOL'; // otherwise " inf" is parsed as ("in", "f")
\s*"/"[<=]"=>"\s*               return 'RELATION';
\s*[<=]"=>"\s*                  return 'RELATION';
\s*"=^o"\s*                     return 'RELATION';
\s*"/"?("in"|"<-"|"->"|[lg]"e"|"="[<>]|[<>~]?"="|[<>]|"approx")\s* return 'RELATION';
\s*("ne")\s*                    return 'RELATION';
\s*".="\s*                      return 'RELATION';
\s*("/"?"su"[bp]"e"?)\s*        return 'RELATION';
(alpha|beta|[Gg]amma|[Dd]elta|(var)?epsilon|(z||(T|(var)?t)h)eta|iota|(var)?kappa|[Ll]ambda|[mn]u|[Xx]i|omicron|(P|(var)?p)i|(var)?rho|(S|(var)?s)igma|tau|[Uu]psilon|((P|(var)?p)|c)hi|[Pp]si|[Oo]mega) return 'SYMBOL';
("undefined"|"UDE"|"ude")       return 'UNDEFINED';
\s+("UDE")                      return 'SPCUNDEFINED';
("DNE"|"dne")                   return 'DOESNOTEXIST';
\s+("dne")                      return 'SPCDOESNOTEXIST';
\s*("U")\s*                     return 'SETOP';
\s*("cap")\s*                   return 'SETOP';
\s*("d"[A-Za-z.])               return 'DX';
\s*("pm"|"+/-")\s*              return 'PM';
\s*"-/+"\s*                     return 'MP';
\s*"+"\s*                       return '+';
\s*"-"\s*                       return '-';
\s*"*"\s*                       return '*';
\s*"/"\s*                       return '/';
\s*"^"([>_^~.]|"->")            return 'SUPDECORATION';
\s*"^"\s*                       return '^';
\s*[_]\s*                       return '_';
\s*","\s*                       return ',';
\s*"!"                          return '!';
\s*"?"                          return '?';
\s*"&"\s*                       return '&';
\s*"\\"\s*                      return 'BACKSLASH';
("CC"|"QQ"|"NN"|"RR"|"Z"[-Z+]|"{}") return 'SYMBOL';
/*"||"                             return '||';*/
[[({]"."?\s*                    return 'LBRACKET';
"|."\s*                         return 'LPIPE';
\s*"."?[\])}]                   return 'RBRACKET';
\s*".|"                         return 'RPIPE';
"|"                             return '|';
("...")                         return 'SYMBOL';
[A-Za-z]                        return 'VARIABLE';
/*"\\"[A-Za-z]+" "                return 'MVARIABLE';*/
/*"``"([^`]|"\\`")+"``"           return 'TEXTRM';*/
<<EOF>>                         return 'EOF';
\s+                             return 'SPC';
/* ignore unrecognized characters, but log them */
.                               if (yy.unrecognized) yy.unrecognized(yytext, yylloc);

/lex

%token 'LATEX'
%token LBRACKET RBRACKET '|' '?'
%token INT DX SUPDECORATION IDX
%token EOF NUMBER VARIABLE SYMBOL UNDEFINED DOESNOTEXIST SPC MISSINGVALUE INPUT
%right ','
%right 'RELATION'
%right 'SETOP'
%left 'BACKSLASH'
%left '&'
%left 'BINARY-PM' 'BINARY-MINUS' 'BINARY-PLUS'
%left 'LIM' 'BIGOP'
%left '/' '*' 'IMPLICIT-MULT'
%right 'ROOT'
%left 'FUNC' 'FUNCSUP' 'FUNCSUB' '^' '-U' 'STRONG-IMPLICIT-MULT' 'PM' 'MP'
%right '^' '_'
%left 'PRIME' SUPDECORATION '!'
%right '_<' '_^<' '^_<' '^<'
%token 'TOP'

%start expressions

%%

/* enable EBNF grammar syntax */
%ebnf

/* dummy rule to force grammar to include some terminals */
dummy
    : '|'
;

atom
    : NUMBER
    { $$ = yy.number(yytext[0] == '.' ? ('0' + yytext) : yytext, @1); }
    | VARIABLE
    { $$ = yy.variable(yytext, @1); }
    | INPUT
    { $$ = yy.input(yytext.substr(2), @1); }
    | LATEX
    { $$ = yy.latex(yytext, @1); }
    | SYMBOL
    { $$ = yy.symbol(yytext, @1); }
    | UNDEFINED
    { $$ = yy.symbol('undefined', @1); }
    | DOESNOTEXIST
    { $$ = yy.symbol('doesnotexist', @1); }
    | DX
    { $$ = yy.symbol(yytext, @1); }
    | MISSINGVALUE
    { $$ = yy.symbol(null, @1); }
    | LATEX
    { $$ = yy.latex(null, @1); }
;

signop
    : '-'
    { $$ = '-'; }
    | 'PM'
    { $$ = '\\pm'; }
    | 'MP'
    { $$ = '\\mp'; }
    | '+'
    { $$ = '+'; }
;

primary
    : atom
    | LBRACKET list? PREEND? (RPIPE | RBRACKET) '?'?
    { $$ = yy.bracket($1, $2 === undefined ? null : $2, $4, $5 !== undefined, @1, @2, @4); }
    | LPIPE SPC? list? SPC? PREEND? (RPIPE | RBRACKET) '?'?
    { $$ = yy.bracket($1, $3 === undefined ? null : $3, $6, $7 !== undefined, @1, @3, @6); }
    | integral SPC? list? PREEND? IDX
    { $$ = yy.bracket($1, $3 === undefined ? null : $3, $5, false, @1, $3 === undefined ? null : @3, @5); }
;

integral
    : INT
    { $$ = 'int'; }
    | INT '_' bigopsub
    { $$ = yy.bigop('int', $3, null, @1, @2, @3, null, null); }
    | INT '^' bigopsup
    { $$ = yy.bigop('int', null, $3, @1, null, null, @2, @3); }
    | INT '_' bigopsub '^' bigopsup
    { $$ = yy.bigop('int', $3, $5, @1, @2, @3, @4, @5); }
    | INT '^' bigopsup '_' bigopsub
    { $$ = yy.bigop('int', $5, $3, @1, @4, @5, @2, @3); }
;

bigopsub
    : bigopsubsup
    | bigopsubsup '_' bigopsub
    { $$ = yy.binop($1, '_', $2, @1, @2, @3); }
;

bigopsup
    : bigopsubsup
    | bigopsubsup '^' bigopsup
    { $$ = yy.binop($1, '^', $2, @1, @2, @3); }
;

bigopsubsup
    : decorated
    | signop bigopsubsup
    { $$ = yy.sign($1, $2, @1, @2); }
    | f operand
    { $$ = yy.func($1, $2, @1, @2); }
;

decorated
    : primary
    | decorated PRIME
    { $$ = yy.postfunc('prime', $1, @2, @1); }
    | decorated '!'
    { $$ = yy.postfunc('factorial', $1, @2, @1); }
    | decorated SUPDECORATION
    { $$ = yy.postfunc($2.trim() === '^>' ? '->' : $2.trim().substr(1), $1, @2, @1); }
;

subsup
    : decorated
    | sub
    | sup
;

sub
    : decorated '_' subscript
    { $$ = yy.binop($1, '_', $3, @1, @2, @3); }
    | sup '_' subscript
    { $$ = yy.binop($1, '_', $3, @1, @2, @3); }
;

sup
    : decorated '^' superscript
    { $$ = yy.binop($1, '^', $3, @1, @2, @3); }
    | sub '^' superscript
    { $$ = yy.binop($1, '^', $3, @1, @2, @3); }
;

subsupoperand
    : decorated
    | signop subsupoperand
    { $$ = yy.sign($1, $2, @1, @2); }
    | f operand
    { $$ = yy.func($1, $2, @1, @2); }
    | subsupoperand decorated
    { $$ = yy.binop($1, '(*)', $2, @1, null, @2); }
;

subscript
    : subsupoperand
    | subsupoperand '_' subscript
    { $$ = yy.binop($1, '_', $3, @1, @2, @3); }
;

superscript
    : subsupoperand
    | subsupoperand '^' superscript
    { $$ = yy.binop($1, '^', $3, @1, @2, @3); }
;

func
    : subsup
    | f operand
    { $$ = yy.func($1, $2, @1, @2); }
    | f SPC operand
    { $$ = yy.func($1, $3, @1, @3); }
    | f PREEND
    { $$ = yy.func($1, yy.symbol(null, null), @1, null); }
    | f SPC PREEND
    { $$ = yy.func($1, yy.symbol(null, null), @1, null); }
    | lpsexp
;

f
    : FUNC
    { $$ = $1.trim(); }
    | FUNCSUB
    { $$ = $1.trim(); }
    | FUNCSUP
    { $$ = $1.trim(); }
    | FUNCSUB '_' sign
    { $$ = yy.binop($1.trim(), '_', $3, @1, @2, @3); }
    | FUNCSUP '^' sign
    { $$ = yy.binop($1.trim(), '^', $3, @1, @2, @3); }
;

operand
    : func
    | operand subsup
    { $$ = yy.binop($1, '(*)', $2, @1, null, @2); }
    | signop operand
    { $$ = yy.sign($1, $2, @1, @2); }
;

lpsexp
    : lps (lpsexp|mult)
    { $$ = yy.func($1, $2, @1, @2); }
    | lps SPC (lpsexp|mult)
    { $$ = yy.func($1, $3, @1, @3); }
    | lps PREEND
    { $$ = yy.func($1, yy.symbol(null, null), @1, null); }
    | lps SPC PREEND
    { $$ = yy.func($1, yy.symbol(null, null), @1, null); }
;

/* limits, products, sum */
lps
    /* product/sum */
    : BIGOP
    { $$ = $1.trim(); }
    | BIGOP '_' bigopsub
    { $$ = yy.bigop($1.trim(), $3, null, @1, @2, @3, null, null); }
    | BIGOP '_' bigopsub RELATION bigopsub
    { $$ = yy.bigop($1.trim(), yy.binop($3, $4.trim(), $5, @3, @4, @5), null, @1, @2, @3, null, null); }
    | BIGOP '^' bigopsup
    { $$ = yy.bigop($1.trim(), null, $3, @1, null, null, @2, @3); }
    | BIGOP '_' bigopsub '^' bigopsup
    { $$ = yy.bigop($1.trim(), $3, $5, @1, @2, @3, @4, @5); }
    | BIGOP '^' bigopsup '_' bigopsub
    { $$ = yy.bigop($1.trim(), $5, $3, @1, @4, @5, @2, @3); }
    | BIGOP '_' bigopsub RELATION bigopsub '^' bigopsup
    { $$ = yy.bigop($1.trim(), yy.binop($3, $4.trim(), $5, @3, @4, @5), $7, @1, @2, @3, @6, @7); }
    | BIGOP '^' bigopsup '_' bigopsub RELATION bigopsub
    { $$ = yy.bigop($1.trim(), yy.binop($5, $6.trim(), $7, @5, @6, @7), $3, @1, @4, @5, @2, @3); }
    /* limits */
    | LIM
    { $$ = $1.trim(); }
    | LIM '_' bigopsub
    { $$ = yy.binop($1.trim(), '_', $3, @1, @2, @3); }
    | LIM '_' bigopsub RELATION bigopsub
    { $$ = yy.binop($1.trim(), '_', yy.binop($3, $4.trim(), $5, @3, @4, @5), @1, @2, @3); }
;

unsignedroot
    : func
    | func ROOT operand
    { $$ = yy.binop($1, 'root', $3, @1, @2, @3); }
;

sign
    : func
    | signop sign
    { $$ = yy.sign($1, $2, @1, @2); }
;

root
    : sign
    | sign ROOT operand
    { $$ = yy.binop($1, 'root', $3, @1, @2, @3); }
;

mult
    : root
    | mult '/' denom
    { $$ = yy.binop($1, '/', $3, @1, @2, @3); }
    | mult '*' root
    { $$ = yy.binop($1, '*', $3, @1, @2, @3); }
    /* implicit multiplication. Note: we never implicitly multiply 'sign' on
     * the right, because that would conflict with 'add' */
    | mult unsignedroot
    { $$ = yy.binop($1, '(*)', $2, @1, null, @2); }
    | mult SPC unsignedroot
    { $$ = yy.binop($1, '(*)', $3, @1, @2, @3); }
;

denom
    : root
    | denom subsup
    { $$ = yy.binop($1, '(*)', $2, @1, null, @2); }
;

add
    : mult
    | add '+' mult
    { $$ = yy.binop($1, '+', $3, @1, @2, @3); }
    | add '-' mult
    { $$ = yy.binop($1, '-', $3, @1, @2, @3); }
    | add 'PM' mult
    { $$ = yy.binop($1, '\\pm', $3, @1, @2, @3); }
    | add 'MP' mult
    { $$ = yy.binop($1, '\\mp', $3, @1, @2, @3); }
;

col
    : add
    | col '&' add
    { $$ = yy.binop($1, '&', $3, @1, @2, @3); }
;

row
    : col
    | row BACKSLASH col
    { $$ = yy.binop($1, '\\', $3, @1, @2, @3); }
;

setop
    : row
    | setop SETOP row
    { $$ = yy.binop($1, $2.trim(), $3, @1, @2, @3); }
;

rel
    : setop
    | rel RELATION setop
    { $$ = yy.binop($1, $2.trim(), $3, @1, @2, @3); }
;

list
    : rel
    | list ',' rel
    { $$ = yy.binop($1, ',', $3, @1, @2, @3); }
;

expressions
    : list PREEND? EOF
    {return $1;}
;
