/* eslint-env node, mocha */
"use strict";

var st2func = require('../lib/func').toFunction;
var expect = require("expect.js");

describe('SmartType to Function converter', function () {
    describe('[atomic]', function () {
        it('should convert a variable', function () {
            var f = st2func('x', {variables: ['x']});
            expect(f).to.be.a(Function);
            expect(f({x: 2})).to.equal(2);
        });
        it('should convert a Greek variable', function () {
            var f = st2func('alpha', {variables: ['alpha']});
            expect(f).to.be.a(Function);
            expect(f({alpha: 3})).to.equal(3);
        });
        it('should convert e', function () {
            var f = st2func('e');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.E);
        });
        it('should convert a pi', function () {
            var f = st2func('pi');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.PI);
        });
        it('should convert a number', function () {
            var f = st2func('2', {variables: []});
            expect(f).to.be.a(Function);
            expect(f()).to.equal(2);
        });
    });
    describe('[basic]', function () {
        it('should convert negation', function () {
            var f = st2func('-x', {variables: ['x']});
            expect(f).to.be.a(Function);
            expect(f({x: 2})).to.equal(-2);
        });
        it('should convert positive sign', function () {
            var f = st2func('+x', {variables: ['x']});
            expect(f).to.be.a(Function);
            expect(f({x: 2})).to.equal(2);
        });
        it('should convert addition', function () {
            var f = st2func('x+y', {variables: ['x', 'y']});
            expect(f).to.be.a(Function);
            expect(f({x: 2, y: 3})).to.equal(5);
        });
        it('should convert subtraction', function () {
            var f = st2func('x-y', {variables: ['x', 'y']});
            expect(f).to.be.a(Function);
            expect(f({x: 2, y: 3})).to.equal(-1);
        });
        it('should convert implicit multiplication', function () {
            var f = st2func('xy', {variables: ['x', 'y']});
            expect(f).to.be.a(Function);
            expect(f({x: 2, y: 3})).to.equal(6);
        });
        it('should convert explicit multiplication', function () {
            var f = st2func('x*y', {variables: ['x', 'y']});
            expect(f).to.be.a(Function);
            expect(f({x: 4, y: 3})).to.equal(12);
        });
        it('should convert division', function () {
            var f = st2func('x/y', {variables: ['x', 'y']});
            expect(f).to.be.a(Function);
            expect(f({x: 6, y: 2})).to.equal(3);
        });
        it('should convert brackets', function () {
            var f = st2func('(3)');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(3);
        });
        it('should convert absolute brackets', function () {
            var f = st2func('|-2|');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(2);
        });
    });
    describe('[functions]', function () {
        it('should convert abs', function () {
            var f = st2func('abs -2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(2);
        });
        it('should convert ceil', function () {
            var f = st2func('ceil 2.2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(3);
        });
        it('should convert floor', function () {
            var f = st2func('floor 2.2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(2);
        });
        it('should convert sqrt', function () {
            var f = st2func('sqrt 2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.sqrt(2));
        });
        it('should convert abs', function () {
            var f = st2func('abs -2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(2);
        });
        it('should convert ln', function () {
            var f = st2func('ln 2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.log(2));
        });
        it('should convert log_10', function () {
            var f = st2func('log_10 2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.log(2)/Math.LN10);
        });
        it('should convert sin', function () {
            var f = st2func('sin 2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.sin(2));
        });
        it('should convert cos', function () {
            var f = st2func('cos 2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.cos(2));
        });
        it('should convert tan', function () {
            var f = st2func('tan 2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.tan(2));
        });
        it('should convert arcsin', function () {
            var f = st2func('arcsin 0.5');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.asin(0.5));
        });
        it('should convert arccos', function () {
            var f = st2func('arccos 0.5');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.acos(0.5));
        });
        it('should convert arctan', function () {
            var f = st2func('arctan 0.5');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(Math.atan(0.5));
        });
        it('should convert min', function () {
            var f = st2func('min(1,2,3)');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(1);
        });
        it('should convert max', function () {
            var f = st2func('max(1,2,3)');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(3);
        });
    });
    describe('[bigop]', function () {
        it('should convert sum', function () {
            var f = st2func('sum_n=1^3 n^2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(14);
        });
        it('should convert prod', function () {
            var f = st2func('prod_n=1^3 n^2');
            expect(f).to.be.a(Function);
            expect(f()).to.equal(36);
        });
    });
});
