/* eslint-env node, mocha */
"use strict";

var st2ltx = require('../lib/latex').toLaTeX;
var expect = require("expect.js");

describe('SmartType to LaTeX converter', function () {
    describe('[atomic]', function () {
        it('should convert a variable', function () {
            expect(st2ltx('x')).to.eql({latex: 'x', errors: []});
        });
        it('should convert Greek letters', function () {
            expect(st2ltx('alpha')).to.eql({latex: '\\alpha ', errors: []});
        });
        it('should convert dx', function () {
            expect(st2ltx('dx')).to.eql({latex: '\\mathrm{d}x', errors: []});
            expect(st2ltx('d.')).to.eql({latex: '\\mathrm{d}', errors: []});
        });
        it('should convert infinity', function () {
            expect(st2ltx('inf')).to.eql({latex: '\\infty ', errors: []});
        });
    });
    describe('[basic]', function () {
        it('should convert addition', function () {
            expect(st2ltx('x+y')).to.eql({latex: 'x+y', errors: []});
            expect(st2ltx('xpmy')).to.eql({latex: 'x\\pm y', errors: []});
            expect(st2ltx('w-xpmy+z')).to.eql({latex: 'w-x\\pm y+z', errors: []});
        });
        it('should convert subscripts', function () {
            expect(st2ltx('x_y')).to.eql({latex: 'x_{y}', errors: []});
        });
        it('should convert superscripts', function () {
            expect(st2ltx('x^y')).to.eql({latex: 'x^{y}', errors: []});
        });
        it('should convert text decorations', function () {
            expect(st2ltx('x^>')).to.eql({latex: '\\vec{x}', errors: []});
            expect(st2ltx('x^^')).to.eql({latex: '\\hat{x}', errors: []});
            expect(st2ltx('x^_')).to.eql({latex: '\\bar{x}', errors: []});
            expect(st2ltx('x^.')).to.eql({latex: '\\dot{x}', errors: []});
            expect(st2ltx('x^~')).to.eql({latex: '\\tilde{x}', errors: []});
        });
        it('should convert postfix operators', function () {
            expect(st2ltx('x!')).to.eql({latex: 'x!', errors: []});
            expect(st2ltx("x'")).to.eql({latex: '{x^{\\prime}}', errors: []});
        });
        it('should convert multiplication', function () {
            expect(st2ltx('xy')).to.eql({latex: 'xy', errors: []});
            expect(st2ltx('x*y')).to.eql({latex: 'xy', errors: []});
            expect(st2ltx('1*2')).to.eql({latex: '1\\cdot 2', errors: []});
        });
        it('should convert division', function () {
            expect(st2ltx('x/y')).to.eql({latex: '\\frac{x}{y}', errors: []});
        });
        it('should convert set operations', function () {
            expect(st2ltx('AUB')).to.eql({latex: 'A\\cup B', errors: []});
            expect(st2ltx('AcapB')).to.eql({latex: 'A\\cap B', errors: []});
        });
        it('should convert regular relations', function () {
            expect(st2ltx('x=y')).to.eql({latex: 'x=y', errors: []});
            expect(st2ltx('x<y')).to.eql({latex: 'x<y', errors: []});
            expect(st2ltx('x>y')).to.eql({latex: 'x>y', errors: []});
            expect(st2ltx('x<=y')).to.eql({latex: 'x\\leq y', errors: []});
            expect(st2ltx('x=<y')).to.eql({latex: 'x\\leq y', errors: []});
            expect(st2ltx('xley')).to.eql({latex: 'x\\leq y', errors: []});
            expect(st2ltx('x>=y')).to.eql({latex: 'x\\geq y', errors: []});
            expect(st2ltx('x=>y')).to.eql({latex: 'x\\geq y', errors: []});
            expect(st2ltx('xgey')).to.eql({latex: 'x\\geq y', errors: []});
            expect(st2ltx('x~=y')).to.eql({latex: 'x\\cong y', errors: []});
            expect(st2ltx('xapproxy')).to.eql({latex: 'x\\approx y', errors: []});
            expect(st2ltx('x.=y')).to.eql({latex: 'x\\doteq y', errors: []});
            expect(st2ltx('x=^oy')).to.eql({latex: 'x\\circeq y', errors: []});
        });
        it('should convert logic relations', function () {
            expect(st2ltx('x==>y')).to.eql({latex: 'x\\Rightarrow y', errors: []});
            expect(st2ltx('x<=>y')).to.eql({latex: 'x\\Leftrightarrow y', errors: []});
        });
        it('should convert set relations', function () {
            expect(st2ltx('x inA')).to.eql({latex: 'x\\in A', errors: []});
            expect(st2ltx('AsubB')).to.eql({latex: 'A\\subset B', errors: []});
            expect(st2ltx('AsupB')).to.eql({latex: 'A\\supset B', errors: []});
            expect(st2ltx('AsubeB')).to.eql({latex: 'A\\subseteq B', errors: []});
            expect(st2ltx('AsupeB')).to.eql({latex: 'A\\supseteq B', errors: []});
        });
        it('should convert negated regular relations', function () {
            expect(st2ltx('x/=y')).to.eql({latex: 'x\\neq y', errors: []});
            expect(st2ltx('x/<y')).to.eql({latex: 'x\\not<y', errors: []});
            expect(st2ltx('x/>y')).to.eql({latex: 'x\\not>y', errors: []});
            expect(st2ltx('x/<=y')).to.eql({latex: 'x\\nleq y', errors: []});
            expect(st2ltx('x/=<y')).to.eql({latex: 'x\\nleq y', errors: []});
            expect(st2ltx('x/ley')).to.eql({latex: 'x\\nleq y', errors: []});
            expect(st2ltx('x/>=y')).to.eql({latex: 'x\\ngeq y', errors: []});
            expect(st2ltx('x/=>y')).to.eql({latex: 'x\\ngeq y', errors: []});
            expect(st2ltx('x/gey')).to.eql({latex: 'x\\ngeq y', errors: []});
            expect(st2ltx('x/~=y')).to.eql({latex: 'x\\ncong y', errors: []});
            expect(st2ltx('x/approxy')).to.eql({latex: 'x\\not\\approx y', errors: []});
        });
        it('should convert negated logic relations', function () {
            expect(st2ltx('x/==>y')).to.eql({latex: 'x\\nRightarrow y', errors: []});
            expect(st2ltx('x/<=>y')).to.eql({latex: 'x\\nLeftrightarrow y', errors: []});
        });
        it('should convert negated set relations', function () {
            expect(st2ltx('x/inA')).to.eql({latex: 'x\\not\\in A', errors: []});
            expect(st2ltx('A/subB')).to.eql({latex: 'A\\not\\subset B', errors: []});
            expect(st2ltx('A/supB')).to.eql({latex: 'A\\not\\supset B', errors: []});
            expect(st2ltx('A/subeB')).to.eql({latex: 'A\\not\\subseteq B', errors: []});
            expect(st2ltx('A/supeB')).to.eql({latex: 'A\\not\\supseteq B', errors: []});
        });
        it('should convert big operators', function () {
            expect(st2ltx('sum x')).to.eql({latex: '\\sum x', errors: []});
            expect(st2ltx('sum_D x')).to.eql({latex: '\\sum_{D}x', errors: []});
            expect(st2ltx('sum^D x')).to.eql({latex: '\\sum^{D}x', errors: []});
            expect(st2ltx('sum_x=0^infx')).to.eql({latex: '\\sum_{x=0}^{\\infty }x', errors: []});
        });
        it('should convert integrals', function () {
            expect(st2ltx('intxdx')).to.eql({latex: '\\int x\\;\\mathrm{d}x', errors: []});
            expect(st2ltx('int_D xdx')).to.eql({latex: '\\int_{D}x\\;\\mathrm{d}x', errors: []});
            expect(st2ltx('int^D xdx')).to.eql({latex: '\\int^{D}x\\;\\mathrm{d}x', errors: []});
            expect(st2ltx('int_0^1 xdx')).to.eql({latex: '\\int_{0}^{1}x\\;\\mathrm{d}x', errors: []});
        });
        it('should convert brackets', function () {
            expect(st2ltx('(x)')).to.eql({latex: '\\left(x\\right)', errors: []});
            expect(st2ltx('{x}')).to.eql({latex: '\\left\\{x\\right\\}', errors: []});
            expect(st2ltx('[x]')).to.eql({latex: '\\left[x\\right]', errors: []});
            expect(st2ltx('|x|')).to.eql({latex: '\\left|x\\right|', errors: []});
        });
        it('should convert hidden brackets', function () {
            expect(st2ltx('(.x)')).to.eql({latex: '\\left.x\\right)', errors: []});
            expect(st2ltx('(x.)')).to.eql({latex: '\\left(x\\right.', errors: []});
            expect(st2ltx('{.x}')).to.eql({latex: '\\left.x\\right\\}', errors: []});
            expect(st2ltx('{x.}')).to.eql({latex: '\\left\\{x\\right.', errors: []});
            expect(st2ltx('|.x|')).to.eql({latex: '\\left.x\\right|', errors: []});
            expect(st2ltx('|x.|')).to.eql({latex: '\\left|x\\right.', errors: []});
        });
    });
    describe('[matrix]', function () {
        it('should convert matrices', function () {
            expect(st2ltx('x & y & z \\ a & b & c')).to.eql({latex: '\\begin{matrix} x & y & z \\\\ a & b & c \\end{matrix}', errors: []});
        });
    });
    describe('[prime]', function () {
        it('should convert multiple primes', function () {
            expect(st2ltx("x''")).to.eql({latex: '{x^{\\prime\\prime}}', errors: []});
        });
    });
    describe('[error handling]', function () {
        it('should show missing values', function () {
            expect(st2ltx('*', {dispmissing: true})).to.eql({latex: '\\Box \\cdot \\Box ', errors: []});
        });
    });
});
