/* eslint-env node, mocha */
"use strict";

var smarttype = require('../lib/smarttype').parser;
var expect = require("expect.js");
var lexer = smarttype.lexer;

var unrecognized = [];

var yy = {
    unrecognized: function (token, location) {
        unrecognized.push({token: token, location: location});
    }
};

var oldyy = lexer.yy;

var expectLexer = function (str)
{
    lexer.setInput(str);
    unrecognized = [];
    var expectedUnrecognized = [];
    return {
        token: function (type, text)
        {
            if (type === 'RBRACKET' || type === 'RPIPE' || type === 'IDX')
            {
                expect(smarttype.terminals_[lexer.lex()]).to.be('PREEND');
            }
            expect(smarttype.terminals_[lexer.lex()]).to.be(type);
            expect(lexer.yytext).to.be(text);
            return this;
        },
        missingvalue: function ()
        {
            expect(smarttype.terminals_[lexer.lex()]).to.be('MISSINGVALUE');
            return this;
        },
        unrecognized: function (token, location)
        {
            var x = {token: token, location: location};
            expectedUnrecognized.push(x);
            expect(unrecognized).to.have.length(expectedUnrecognized.length);
            expect(unrecognized[unrecognized.length - 1]).to.eql(x);
            return this;
        },
        eof: function ()
        {
            expect(smarttype.terminals_[lexer.lex()]).to.be('PREEND');
            expect(smarttype.terminals_[lexer.lex()]).to.be('EOF');
            expect(lexer.yytext).to.be('');
            expect(unrecognized).to.have.length(expectedUnrecognized.length);
            return this;
        },
        singleToken: function (type)
        {
            return this.token(type, str).eof();
        }
    };
};

describe('SmartType Lexer', function () {
    beforeEach(function () {
        lexer.yy = yy;
    });
    afterEach(function () {
        lexer.yy = oldyy;
    });
    describe('[numbers]', function () {
        it('should recognize integers', function () {
            expectLexer('1234').singleToken('NUMBER');
        });
        it('should recognize floats', function () {
            expectLexer('3.14159265').singleToken('NUMBER');
        });
    });
    describe('[functions]', function () {
        it('should recognize regular trig functions', function () {
            expectLexer('sin').token('FUNCSUP', 'sin').missingvalue().eof();
            expectLexer('cos').token('FUNCSUP', 'cos').missingvalue().eof();
            expectLexer('tan').token('FUNCSUP', 'tan').missingvalue().eof();
            expectLexer('sec').token('FUNCSUP', 'sec').missingvalue().eof();
            expectLexer('csc').token('FUNCSUP', 'csc').missingvalue().eof();
            expectLexer('cot').token('FUNCSUP', 'cot').missingvalue().eof();
        });
        it('should recognize inverse trig functions', function () {
            expectLexer('arcsin').token('FUNCSUP', 'arcsin').missingvalue().eof();
            expectLexer('arccos').token('FUNCSUP', 'arccos').missingvalue().eof();
            expectLexer('arctan').token('FUNCSUP', 'arctan').missingvalue().eof();
            expectLexer('arcsec').token('FUNCSUP', 'arcsec').missingvalue().eof();
            expectLexer('arccsc').token('FUNCSUP', 'arccsc').missingvalue().eof();
            expectLexer('arccot').token('FUNCSUP', 'arccot').missingvalue().eof();
        });
        it('should recognize hyperbolic trig functions', function () {
            expectLexer('sinh').token('FUNCSUP', 'sinh').missingvalue().eof();
            expectLexer('cosh').token('FUNCSUP', 'cosh').missingvalue().eof();
            expectLexer('tanh').token('FUNCSUP', 'tanh').missingvalue().eof();
            expectLexer('sech').token('FUNCSUP', 'sech').missingvalue().eof();
            expectLexer('csch').token('FUNCSUP', 'csch').missingvalue().eof();
            expectLexer('coth').token('FUNCSUP', 'coth').missingvalue().eof();
        });
        it('should recognize inverse hyperbolic trig functions', function () {
            expectLexer('arcsinh').token('FUNCSUP', 'arcsinh').missingvalue().eof();
            expectLexer('arccosh').token('FUNCSUP', 'arccosh').missingvalue().eof();
            expectLexer('arctanh').token('FUNCSUP', 'arctanh').missingvalue().eof();
            expectLexer('arcsech').token('FUNCSUP', 'arcsech').missingvalue().eof();
            expectLexer('arccsch').token('FUNCSUP', 'arccsch').missingvalue().eof();
            expectLexer('arccoth').token('FUNCSUP', 'arccoth').missingvalue().eof();
        });
        it('should recognize other functions', function () {
            expectLexer('log').token('FUNCSUB', 'log').missingvalue().eof();
            expectLexer('ln').token('FUNC', 'ln').missingvalue().eof();
            expectLexer('ceil').token('FUNC', 'ceil').missingvalue().eof();
            expectLexer('floor').token('FUNC', 'floor').missingvalue().eof();
            expectLexer('sqrt').token('FUNC', 'sqrt').missingvalue().eof();
            expectLexer('abs').token('FUNC', 'abs').missingvalue().eof();
            expectLexer('exp').token('FUNC', 'exp').missingvalue().eof();
            expectLexer('min').token('FUNC', 'min').missingvalue().eof();
            expectLexer('max').token('FUNC', 'max').missingvalue().eof();
        });
    });
    describe('[operators]', function () {
        it('should recognize post-operators', function () {
            expectLexer('!').missingvalue().token('!', '!').eof();
            expectLexer("'").missingvalue().token('PRIME', "'").eof();
        });
        it('should recognize text decorations', function () {
            expectLexer('^>').missingvalue().token('SUPDECORATION', '^>').eof();
            expectLexer('^->').missingvalue().token('SUPDECORATION', '^->').eof();
            expectLexer('^_').missingvalue().token('SUPDECORATION', '^_').eof();
            expectLexer('^^').missingvalue().token('SUPDECORATION', '^^').eof();
            expectLexer('^~').missingvalue().token('SUPDECORATION', '^~').eof();
            expectLexer('^.').missingvalue().token('SUPDECORATION', '^.').eof();
        });
        it('should recognize big operators and limits', function () {
            expectLexer('sum').token('BIGOP', 'sum').missingvalue().eof();
            expectLexer('prod').token('BIGOP', 'prod').missingvalue().eof();
            expectLexer('lim').token('LIM', 'lim').missingvalue().eof();
        });
        it('should recognize set operators', function () {
            expectLexer('U').missingvalue().token('SETOP', 'U').missingvalue().eof();
            expectLexer('cap').missingvalue().token('SETOP', 'cap').missingvalue().eof();
        });
        it('should recognize binary operators', function () {
            expectLexer('+').token('+', '+').missingvalue().eof();
            expectLexer('-').token('-', '-').missingvalue().eof();
            expectLexer('*').missingvalue().token('*', '*').missingvalue().eof();
            expectLexer('/').missingvalue().token('/', '/').missingvalue().eof();
            expectLexer('pm').token('PM', 'pm').missingvalue().eof();
            expectLexer('+/-').token('PM', '+/-').missingvalue().eof();
            expectLexer('-/+').token('MP', '-/+').missingvalue().eof();
            expectLexer('root').missingvalue().token('ROOT', 'root').missingvalue().eof();
            expectLexer('^').missingvalue().token('^', '^').missingvalue().eof();
            expectLexer('_').missingvalue().token('_', '_').missingvalue().eof();
        });
        it('should recognize separators', function () {
            expectLexer(',').missingvalue().token(',', ',').missingvalue().eof();
            expectLexer('&').missingvalue().token('&', '&').missingvalue().eof();
            expectLexer('\\').missingvalue().token('BACKSLASH', '\\').missingvalue().eof();
        });
    });
    describe('[symbols]', function () {
        it('should recognize infinity', function () {
            expectLexer('inf').singleToken('SYMBOL');
        });
        it('should recognize lower case Greek letters', function () {
            expectLexer('alpha').singleToken('SYMBOL');
            expectLexer('beta').singleToken('SYMBOL');
            expectLexer('gamma').singleToken('SYMBOL');
            expectLexer('delta').singleToken('SYMBOL');
            expectLexer('epsilon').singleToken('SYMBOL');
            expectLexer('zeta').singleToken('SYMBOL');
            expectLexer('eta').singleToken('SYMBOL');
            expectLexer('theta').singleToken('SYMBOL');
            expectLexer('iota').singleToken('SYMBOL');
            expectLexer('kappa').singleToken('SYMBOL');
            expectLexer('lambda').singleToken('SYMBOL');
            expectLexer('mu').singleToken('SYMBOL');
            expectLexer('nu').singleToken('SYMBOL');
            expectLexer('xi').singleToken('SYMBOL');
            expectLexer('omicron').singleToken('SYMBOL');
            expectLexer('pi').singleToken('SYMBOL');
            expectLexer('rho').singleToken('SYMBOL');
            expectLexer('sigma').singleToken('SYMBOL');
            expectLexer('tau').singleToken('SYMBOL');
            expectLexer('upsilon').singleToken('SYMBOL');
            expectLexer('phi').singleToken('SYMBOL');
            expectLexer('chi').singleToken('SYMBOL');
            expectLexer('psi').singleToken('SYMBOL');
            expectLexer('omega').singleToken('SYMBOL');
        });
        it('should recognize upper case Greek letters', function () {
            expectLexer('Gamma').singleToken('SYMBOL');
            expectLexer('Delta').singleToken('SYMBOL');
            expectLexer('Theta').singleToken('SYMBOL');
            expectLexer('Lambda').singleToken('SYMBOL');
            expectLexer('Xi').singleToken('SYMBOL');
            expectLexer('Pi').singleToken('SYMBOL');
            expectLexer('Sigma').singleToken('SYMBOL');
            expectLexer('Upsilon').singleToken('SYMBOL');
            expectLexer('Phi').singleToken('SYMBOL');
            expectLexer('Psi').singleToken('SYMBOL');
            expectLexer('Omega').singleToken('SYMBOL');
        });
        it('should recognize alternate form Greek letters', function () {
            expectLexer('varepsilon').singleToken('SYMBOL');
            expectLexer('vartheta').singleToken('SYMBOL');
            expectLexer('varkappa').singleToken('SYMBOL');
            expectLexer('varpi').singleToken('SYMBOL');
            expectLexer('varsigma').singleToken('SYMBOL');
            expectLexer('varphi').singleToken('SYMBOL');
        });
        it('should recognize common sets', function () {
            expectLexer('CC').singleToken('SYMBOL');
            expectLexer('QQ').singleToken('SYMBOL');
            expectLexer('NN').singleToken('SYMBOL');
            expectLexer('RR').singleToken('SYMBOL');
            expectLexer('ZZ').singleToken('SYMBOL');
            expectLexer('Z+').singleToken('SYMBOL');
            expectLexer('Z-').singleToken('SYMBOL');
            expectLexer('{}').singleToken('SYMBOL');
        });
        it('should recognize dots', function () {
            expectLexer('...').singleToken('SYMBOL');
        });
        it('should recognize dx', function () {
            expectLexer('dx').singleToken('DX');
            expectLexer('d.').singleToken('DX');
        });
        it('should fetch longest token after a space', function () {
            expectLexer(' inf').token('SPC', ' ').token('SYMBOL', 'inf').eof();
            expectLexer(' int').token('SPC', ' ').token('INT', 'int').missingvalue().token('IDX', '').eof();
            expectLexer(' delta').token('SPC', ' ').token('SYMBOL', 'delta').eof();
            expectLexer(' Upsilon').token('SPC', ' ').token('SYMBOL', 'Upsilon').eof();
            expectLexer(' UDE').token('SPC', ' ').token('UNDEFINED', 'UDE').eof();
            expectLexer(' dne').token('SPC', ' ').token('DOESNOTEXIST', 'dne').eof();
        });
        it('should recognize special symbols', function () {
            expectLexer('undefined').singleToken('UNDEFINED');
            expectLexer('ude').singleToken('UNDEFINED');
            expectLexer('UDE').singleToken('UNDEFINED');
            expectLexer('DNE').singleToken('DOESNOTEXIST');
            expectLexer('dne').singleToken('DOESNOTEXIST');
        });
    });
    describe('[relations]', function () {
        it('should recognize regular relations', function () {
            expectLexer('=').missingvalue().token('RELATION', '=').missingvalue().eof();
            expectLexer('<').missingvalue().token('RELATION', '<').missingvalue().eof();
            expectLexer('>').missingvalue().token('RELATION', '>').missingvalue().eof();
            expectLexer('<=').missingvalue().token('RELATION', '<=').missingvalue().eof();
            expectLexer('>=').missingvalue().token('RELATION', '>=').missingvalue().eof();
            expectLexer('=<').missingvalue().token('RELATION', '=<').missingvalue().eof();
            expectLexer('=>').missingvalue().token('RELATION', '=>').missingvalue().eof();
            expectLexer('le').missingvalue().token('RELATION', 'le').missingvalue().eof();
            expectLexer('ge').missingvalue().token('RELATION', 'ge').missingvalue().eof();
            expectLexer('approx').missingvalue().token('RELATION', 'approx').missingvalue().eof();
            expectLexer('.=').missingvalue().token('RELATION', '.=').missingvalue().eof();
            expectLexer('=^o').missingvalue().token('RELATION', '=^o').missingvalue().eof();
        });
        it('should recognize set relations', function () {
            expectLexer('in').missingvalue().token('RELATION', 'in').missingvalue().eof();
            expectLexer('sub').missingvalue().token('RELATION', 'sub').missingvalue().eof();
            expectLexer('sup').missingvalue().token('RELATION', 'sup').missingvalue().eof();
            expectLexer('sube').missingvalue().token('RELATION', 'sube').missingvalue().eof();
            expectLexer('supe').missingvalue().token('RELATION', 'supe').missingvalue().eof();

        });
        it('should recognize arrows', function () {
            expectLexer('==>').missingvalue().token('RELATION', '==>').missingvalue().eof();
            expectLexer('<=>').missingvalue().token('RELATION', '<=>').missingvalue().eof();
            expectLexer('->').missingvalue().token('RELATION', '->').missingvalue().eof();
            expectLexer('<-').missingvalue().token('RELATION', '<-').missingvalue().eof();
        });
        it('should recognize negated regular relations', function () {
            expectLexer('/=').missingvalue().token('RELATION', '/=').missingvalue().eof();
            expectLexer('/<').missingvalue().token('RELATION', '/<').missingvalue().eof();
            expectLexer('/>').missingvalue().token('RELATION', '/>').missingvalue().eof();
            expectLexer('/<=').missingvalue().token('RELATION', '/<=').missingvalue().eof();
            expectLexer('/>=').missingvalue().token('RELATION', '/>=').missingvalue().eof();
            expectLexer('/=<').missingvalue().token('RELATION', '/=<').missingvalue().eof();
            expectLexer('/=>').missingvalue().token('RELATION', '/=>').missingvalue().eof();
            expectLexer('/le').missingvalue().token('RELATION', '/le').missingvalue().eof();
            expectLexer('/ge').missingvalue().token('RELATION', '/ge').missingvalue().eof();
            expectLexer('/approx').missingvalue().token('RELATION', '/approx').missingvalue().eof();
            expectLexer('ne').missingvalue().token('RELATION', 'ne').missingvalue().eof();
        });
        it('should recognize negated set relations', function () {
            expectLexer('/in').missingvalue().token('RELATION', '/in').missingvalue().eof();
            expectLexer('/sub').missingvalue().token('RELATION', '/sub').missingvalue().eof();
            expectLexer('/sup').missingvalue().token('RELATION', '/sup').missingvalue().eof();
            expectLexer('/sube').missingvalue().token('RELATION', '/sube').missingvalue().eof();
            expectLexer('/supe').missingvalue().token('RELATION', '/supe').missingvalue().eof();

        });
        it('should recognize negated arrows', function () {
            expectLexer('/->').missingvalue().token('RELATION', '/->').missingvalue().eof();
            expectLexer('/<-').missingvalue().token('RELATION', '/<-').missingvalue().eof();
        });
    });
    describe('[brackets]', function () {
        it('should recognize normal parentheses and brackets', function () {
            expectLexer('()').token('LBRACKET', '(').missingvalue().token('RBRACKET', ')').eof();
            expectLexer('{x}').token('LBRACKET', '{').token('VARIABLE', 'x').token('RBRACKET', '}').eof();
            expectLexer('[]').token('LBRACKET', '[').missingvalue().token('RBRACKET', ']').eof();
        });
        it('should recognize hidden parentheses and brackets', function () {
            expectLexer('(..)').token('LBRACKET', '(.').missingvalue().token('RBRACKET', '.)').eof();
            expectLexer('{..}').token('LBRACKET', '{.').missingvalue().token('RBRACKET', '.}').eof();
            expectLexer('[..]').token('LBRACKET', '[.').missingvalue().token('RBRACKET', '.]').eof();
        });
        it('should recognize vertical bar brackets', function () {
            expectLexer('||')
                .token('LPIPE', '|')
                .missingvalue()
                .token('RPIPE', '|')
                .eof();
        });
        it('should recognize hidden vertical bar brackets', function () {
            expectLexer('|..|').token('LPIPE', '|.').missingvalue().token('RPIPE', '.|').eof();
        });
        it('should recognize integrals', function () {
            expectLexer('intdx')
                .token('INT', 'int')
                .missingvalue()
                .token('IDX', 'dx')
                .eof();
        });
    });
    describe('[spaces]', function () {
        it('should recognize a space', function () {
            expectLexer(' ').token('SPC', ' ').missingvalue().eof();
        });
        it('should recognize keywords with no spaces between', function () {
            expectLexer('pmsinalpha')
                .token('PM', 'pm')
                .token('FUNCSUP', 'sin')
                .token('SYMBOL', 'alpha')
                .eof();
        });
        it('should not recognize a keyword with spaces in it', function () {
            expectLexer('l n')
                .token('VARIABLE', 'l')
                .token('SPC', ' ')
                .token('VARIABLE', 'n')
                .eof();
        });
    });
    describe('[unrecognized]', function () {
        it('should ignore unrecognized characters', function () {
            expectLexer('l$n')
                .token('VARIABLE', 'l')
                .token('VARIABLE', 'n')
                .unrecognized('$', {first_line: 1, last_line: 1, first_column: 1, last_column: 2})
                .eof();
        });
    });
    describe('[error handling]', function () {
        it('should add missing brackets', function () {
            expectLexer(')').token('LBRACKET', '(').missingvalue().token('RBRACKET', ')').eof();
            expectLexer('}').token('LBRACKET', '{').missingvalue().token('RBRACKET', '}').eof();
            expectLexer(']').token('LBRACKET', '[').missingvalue().token('RBRACKET', ']').eof();
            expectLexer('(').token('LBRACKET', '(').missingvalue().token('RBRACKET', ')').eof();
            expectLexer('{').token('LBRACKET', '{').missingvalue().token('RBRACKET', '}').eof();
            expectLexer('[').token('LBRACKET', '[').missingvalue().token('RBRACKET', ']').eof();
            expectLexer('|').token('LPIPE', '|').missingvalue().token('RPIPE', '|').eof();
            expectLexer('int').token('INT', 'int').missingvalue().token('IDX', '').eof();
            expectLexer('int)')
                .token('LBRACKET', '(')
                .token('INT', 'int')
                .missingvalue()
                .token('IDX', '')
                .token('RBRACKET', ')')
                .eof();
            expectLexer('int)dx')
                .token('LBRACKET', '(')
                .token('INT', 'int')
                .missingvalue()
                .token('IDX', '')
                .token('RBRACKET', ')')
                .token('DX', 'dx')
                .eof();
        });
    });
});
