/* eslint-env node, mocha */
"use strict";

var smarttype = require('../lib/smarttype').parser;
var expect = require("expect.js");

smarttype.trace = console.log.bind(console);
var yy = {
    binop: function (e1, op, e2)
    {
        return {type: 'binop', op: op, left: e1, right: e2};
    },
    bigop: function (op, sub, sup)
    {
        return {type: 'bigop', op: op, sub: sub, sup: sup};
    },
    sign: function (op, e)
    {
        return {type: 'sign', op: op, value: e};
    },
    func: function (name, e)
    {
        return {type: 'function', name: name, value: e};
    },
    postfunc: function (name, e)
    {
        return {type: 'postfunction', name: name, value: e};
    },
    number: function (value)
    {
        return {type: 'number', value: value};
    },
    variable: function (name)
    {
        return {type: 'variable', value: name};
    },
    symbol: function (name)
    {
        return {type: 'symbol', value: name};
    },
    bracket: function (lbracket, value, rbracket, hidden)
    {
        return {type: 'bracket', lbracket: lbracket, value: value, rbracket: rbracket, hidden: !!hidden};
    }
};

var oldyy = smarttype.yy;

describe('SmartType Parser', function () {
    beforeEach(function () {
        smarttype.yy = yy;
    });
    afterEach(function () {
        smarttype.yy = oldyy;
    });
    describe('[atomic]', function () {
        it('should parse a symbol', function () {
            expect(smarttype.parse('alpha')).to.eql(yy.symbol('alpha'));
        });
        it('should parse a variable', function () {
            expect(smarttype.parse('a')).to.eql(yy.variable('a'));
        });
        it('should parse an integer', function () {
            expect(smarttype.parse('42')).to.eql(yy.number('42'));
        });
        it('should parse a decimal number', function () {
            expect(smarttype.parse('3.14159265')).to.eql(yy.number('3.14159265'));
        });
    });
    describe('[basic expressions]', function () {
        it('should parse prime', function () {
            expect(smarttype.parse("x'"))
                .to.eql(yy.postfunc("prime", yy.variable('x')));
        });
        it('should parse factorial', function () {
            expect(smarttype.parse("x!"))
                .to.eql(yy.postfunc("factorial", yy.variable('x')));
        });
        it('should parse vec decoration', function () {
            expect(smarttype.parse("x^>"))
                .to.eql(yy.postfunc("->", yy.variable('x')));
            expect(smarttype.parse("x^->"))
                .to.eql(yy.postfunc("->", yy.variable('x')));
        });
        it('should parse hat decoration', function () {
            expect(smarttype.parse("x^^"))
                .to.eql(yy.postfunc("^", yy.variable('x')));
        });
        it('should parse bar decoration', function () {
            expect(smarttype.parse("x^_"))
                .to.eql(yy.postfunc("_", yy.variable('x')));
        });
        it('should parse tilde decoration', function () {
            expect(smarttype.parse("x^~"))
                .to.eql(yy.postfunc("~", yy.variable('x')));
        });
        it('should parse dot decoration', function () {
            expect(smarttype.parse("x^."))
                .to.eql(yy.postfunc(".", yy.variable('x')));
        });
        it('should parse subscripts', function () {
            expect(smarttype.parse('x_2'))
                .to.eql(yy.binop(yy.variable('x'), '_', yy.number('2')));
        });
        it('should parse superscripts', function () {
            expect(smarttype.parse('x^2'))
                .to.eql(yy.binop(yy.variable('x'), '^', yy.number('2')));
        });
        it('should parse functions', function () {
            expect(smarttype.parse('sinpi'))
                   .to.eql(yy.func('sin', yy.symbol('pi')));
        });
        it('should parse n\'th roots', function () {
            expect(smarttype.parse('3rootx'))
                .to.eql(yy.binop(yy.number('3'), 'root', yy.variable('x')));
        });
        it('should parse signed variables', function () {
            expect(smarttype.parse('-x'))
                .to.eql(yy.sign('-', yy.variable('x')));
        });
        it('should parse a product with no limits', function () {
            expect(smarttype.parse('prodx'))
                .to.eql(yy.func('prod', yy.variable('x')));
        });
        it('should parse a sum with bottom limits', function () {
            expect(smarttype.parse('sum_ZZx'))
                .to.eql(yy.func(yy.bigop('sum', yy.symbol('ZZ'), null), yy.variable('x')));
        });
        it('should parse a sum with limits', function () {
            expect(smarttype.parse('sum_y^zx'))
                .to.eql(yy.func(yy.bigop('sum', yy.variable('y'), yy.variable('z')), yy.variable('x')));
        });
        it('should parse a sum with limits (reversed)', function () {
            expect(smarttype.parse('sum^z_yx'))
                .to.eql(yy.func(yy.bigop('sum', yy.variable('y'), yy.variable('z')), yy.variable('x')));
        });
        it('should parse a sum with a relation on the bottom', function () {
            expect(smarttype.parse('sum_yinZZy'))
                .to.eql(yy.func(yy.bigop('sum', yy.binop(yy.variable('y'), 'in', yy.symbol('ZZ')), null), yy.variable('y')));
        });
        it('should parse a sum with limits', function () {
            expect(smarttype.parse('sum_x=0^infx'))
                .to.eql(yy.func(yy.bigop('sum', yy.binop(yy.variable('x'), '=', yy.number('0')), yy.symbol('inf')), yy.variable('x')));
        });
        it('should parse a sum with negative limits', function () {
            expect(smarttype.parse('sum_x=-6^-2x'))
                .to.eql(yy.func(yy.bigop('sum', yy.binop(yy.variable('x'), '=', yy.sign('-', yy.number('6'))), yy.sign('-', yy.number('2'))), yy.variable('x')));
        });
        it('should parse a limit with no subscript', function () {
            expect(smarttype.parse('limx'))
                   .to.eql(yy.func('lim', yy.variable('x')));
        });
        it('should parse a limit with a subscript', function () {
            expect(smarttype.parse('lim_xx'))
                .to.eql(yy.func(yy.binop('lim', '_', yy.variable('x')), yy.variable('x')));
        });
        it('should parse a limit with a rightarrow', function () {
            expect(smarttype.parse('lim_x->infx'))
                .to.eql(yy.func(yy.binop('lim', '_', yy.binop(yy.variable('x'), '->', yy.symbol('inf'))), yy.variable('x')));
        });
        it('should parse division', function () {
            expect(smarttype.parse('1/2'))
                .to.eql(yy.binop(yy.number('1'), '/', yy.number('2')));
        });
        it('should parse implicit multiplication', function () {
            expect(smarttype.parse('xy'))
                .to.eql(yy.binop(yy.variable('x'), '(*)', yy.variable('y')));
        });
        it('should parse addition', function () {
            expect(smarttype.parse('1+2'))
                .to.eql(yy.binop(yy.number('1'), '+', yy.number('2')));
        });
        it('should parse a matrix column', function () {
            expect(smarttype.parse('1&2'))
                .to.eql(yy.binop(yy.number('1'), '&', yy.number('2')));
        });
        it('should parse a matrix row', function () {
            expect(smarttype.parse('1\\2'))
                   .to.eql(yy.binop(yy.number('1'), '\\', yy.number('2')));
        });
        it('should parse a set operation', function () {
            expect(smarttype.parse('AUB'))
                .to.eql(yy.binop(yy.variable('A'), 'U', yy.variable('B')));
        });
        it('should parse a relation', function () {
            expect(smarttype.parse('1<2'))
                .to.eql(yy.binop(yy.number('1'), '<', yy.number('2')));
        });
        it('should parse a list', function () {
            expect(smarttype.parse('1,2'))
                .to.eql(yy.binop(yy.number('1'), ',', yy.number('2')));
        });
    });
    describe('[bracketed expressions]', function () {
        it('should parse an expression in optional parentheses', function () {
            expect(smarttype.parse('(x)?'))
                .to.eql(yy.bracket('(', yy.variable('x'), ')', true));
        });
        it('should parse an expression in parentheses', function () {
            expect(smarttype.parse('(x)'))
                .to.eql(yy.bracket('(', yy.variable('x'), ')', false));
        });
        it('should allow mixed brackets', function () {
            expect(smarttype.parse('{x]'))
                .to.eql(yy.bracket('{', yy.variable('x'), ']', false));
        });
        it('should parse vertical bars', function () {
            expect(smarttype.parse('|x|'))
                .to.eql(yy.bracket('|', yy.variable('x'), '|', false));
        });
        it('should parse optional bars', function () {
            expect(smarttype.parse('|x|?'))
                .to.eql(yy.bracket('|', yy.variable('x'), '|', true));
        });
        it('should not close parentheses with vertical bars', function () {
            expect(smarttype.parse('(|x|)'))
                .to.eql(yy.bracket('(', yy.bracket('|', yy.variable('x'), '|', false), ')', false));
        });
        it('should implicitly multiply absolute values', function () {
            expect(smarttype.parse('x|x|'))
                .to.eql(yy.binop(yy.variable('x'), '(*)', yy.bracket('|', yy.variable('x'), '|', false), ')', false));
        });
        it('should parse an integral', function () {
            expect(smarttype.parse('intxdx'))
                .to.eql(yy.bracket('int', yy.variable('x'), 'dx', false));
        });
        it('should parse an integral with subscripts', function () {
            expect(smarttype.parse('int_Dxdx'))
                .to.eql(yy.bracket(yy.bigop('int', yy.variable('D'), null), yy.variable('x'), 'dx', false));
        });
        it('should parse an integral with superscripts', function () {
            expect(smarttype.parse('int^Dxdx'))
                .to.eql(yy.bracket(yy.bigop('int', null, yy.variable('D')), yy.variable('x'), 'dx', false));
        });
        it('should parse an integral with sub- and superscripts', function () {
            expect(smarttype.parse('int_0^1xdx'))
                .to.eql(yy.bracket(yy.bigop('int', yy.number(0), yy.number('1')), yy.variable('x'), 'dx', false));
        });
        it('should parse an integral with sub- and superscripts reversed', function () {
            expect(smarttype.parse('int^1_0xdx'))
                .to.eql(yy.bracket(yy.bigop('int', yy.number(0), yy.number('1')), yy.variable('x'), 'dx', false));
        });
        it('should parse an integral with signed sub- and superscripts', function () {
            expect(smarttype.parse('int_-2^-1xdx'))
                .to.eql(yy.bracket(yy.bigop('int', yy.sign('-', yy.number('2')), yy.sign('-', yy.number('1'))), yy.variable('x'), 'dx', false));
        });
        it('should parse a function call within an integral', function () {
            expect(smarttype.parse('intsinxdx'))
                .to.eql(yy.bracket('int', yy.func('sin', yy.variable('x')), 'dx', false));
        });
        it('should parse addition within an integral', function () {
            expect(smarttype.parse('int1+xdx'))
                .to.eql(yy.bracket('int', yy.binop(yy.number('1'), '+', yy.variable('x')), 'dx', false));
        });
        it('should parse vertical bars within an integral', function () {
            expect(smarttype.parse('int|x|dx'))
                .to.eql(yy.bracket('int', yy.bracket('|', yy.variable('x'), '|', false), 'dx', false));
        });
        it('should accept space before dx', function () {
            expect(smarttype.parse('int x dx'))
                .to.eql(yy.bracket('int', yy.variable('x'), 'dx', false));
        });
    });
    describe('[signed expressions]', function () {
        it('should handle mixed binary and unary sign operators', function () {
            expect(smarttype.parse('1+-2'))
                .to.eql(yy.binop(yy.number('1'), '+', yy.sign('-', yy.number('2'))));
        });
    });
    describe('[precedence]', function () {
        it('should not implicitly multiply signed expressions on the right', function () {
            // x-y could be parsed with either "-" as a binary operation
            // (correct), or with "-" as a unary operation with implicit
            // multiplication between x and -y (nonsense)
            expect(smarttype.parse('x-y'))
                .to.eql(yy.binop(yy.variable('x'), '-', yy.variable('y')));
        });
        it('should bind function calls higher than implicit multiplication', function () {
            expect(smarttype.parse('sinasinb'))
                .to.eql(yy.binop(yy.func('sin', yy.variable('a')), '(*)', yy.func('sin', yy.variable('b'))));
        });
        it('should perform implicit multiplication within a function call', function () {
            expect(smarttype.parse('sinab'))
                .to.eql(yy.func('sin', yy.binop(yy.variable('a'), '(*)', yy.variable('b'))));
        });
        it('should not perform implicit multiplication within a function call with a space between', function () {
            expect(smarttype.parse('sina b'))
                .to.eql(yy.binop(yy.func('sin', yy.variable('a')), '(*)', yy.variable('b')));
        });
        it('should bind subscripts higher than function calls', function () {
            expect(smarttype.parse('sinx_2'))
                .to.eql(yy.func('sin', yy.binop(yy.variable('x'), '_', yy.number('2'))));
        });
        it('should bind superscripts higher than function calls', function () {
            expect(smarttype.parse('sinx^2'))
                .to.eql(yy.func('sin', yy.binop(yy.variable('x'), '^', yy.number('2'))));
        });
        it('should bind function calls higher than division', function () {
            expect(smarttype.parse('sin1/x'))
                .to.eql(yy.binop(yy.func('sin', yy.number('1')), '/', yy.variable('x')));
        });
        it('should implicitly multiply terms in the denominator', function () {
            // parse 1/ab as 1/(ab) instead of (1/a)b
            expect(smarttype.parse('1/ab'))
                .to.eql(yy.binop(yy.number('1'), '/', yy.binop(yy.variable('a'), '(*)', yy.variable('b'))));
        });
        it('should not implicitly multiply terms in the denominator with a space between', function () {
            // parse 1/a b as (1/a)b instead of 1/(ab)
            expect(smarttype.parse('1/a b'))
                .to.eql(yy.binop(yy.binop(yy.number('1'), '/', yy.variable('a')), '(*)', yy.variable('b')));
        });
        it('should bind superscripting higher than negation', function () {
            expect(smarttype.parse('-x^2'))
                .to.eql(yy.sign('-', yy.binop(yy.variable('x'), '^', yy.number('2'))));
        });
        it('should bind root higher than multiplication', function () {
            expect(smarttype.parse('2*3root4*5'))
                .to.eql(yy.binop(yy.binop(yy.number('2'), '*', yy.binop(yy.number('3'), 'root', yy.number('4'))), '*', yy.number('5')));
        });
        it('should bind exponent lower than implicit multiplication with a variable', function () {
            expect(smarttype.parse('x^2y'))
                .to.eql(yy.binop(yy.variable('x'), '^', yy.binop(yy.number('2'), '(*)', yy.variable('y'))));
        });
        it('should bind exponent lower than implicit multiplication with a factorial', function () {
            expect(smarttype.parse('x^2y!'))
                .to.eql(yy.binop(yy.variable('x'), '^', yy.binop(yy.number('2'), '(*)', yy.postfunc('factorial', yy.variable('y')))));
        });
        it('should bind exponent higher than implicit multiplication with a function', function () {
            expect(smarttype.parse('x^2sinx'))
                .to.eql(yy.binop(yy.binop(yy.variable('x'), '^', yy.number('2')), '(*)', yy.func('sin', yy.variable('x'))));
        });
        it('should bind division higher than sum', function () {
            expect(smarttype.parse('sum_x=0^inf1/x'))
                .to.eql(yy.func(yy.bigop('sum',
                                         yy.binop(yy.variable('x'), '=', yy.number('0')),
                                         yy.symbol('inf')),
                                yy.binop(yy.number('1'), '/', yy.variable('x'))));
        });
        it('should implicitly multiply dx in the denominator', function () {
            // parse 1/dydx as 1/(dydx) instead of (1/dy)dx
            expect(smarttype.parse('1/dydx'))
                .to.eql(yy.binop(yy.number('1'), '/', yy.binop(yy.symbol('dy'), '(*)', yy.symbol('dx'))));
        });
        it('should not implicitly multiply terms in the denominator with a space between', function () {
            // parse 1/dy dx as (1/dy)dx instead of 1/(dy dx)
            expect(smarttype.parse('1/dy dx'))
                .to.eql(yy.binop(yy.binop(yy.number('1'), '/', yy.symbol('dy')), '(*)', yy.symbol('dx')));
        });
        it.skip('should negate fraction instead of numerator', function () {
            // -1/2 should be parsed as -(1/2) instead of (-1)/2
            expect(smarttype.parse('-1/2'))
                .to.eql(yy.sign('-', yy.binop(yy.number('1'), '/', yy.number('2'))));
        });
    });
    describe('[associativity]', function () {
        it('should right-associate superscripts', function () {
            expect(smarttype.parse('x^2^2'))
                .to.eql(yy.binop(yy.variable('x'), '^', yy.binop(yy.number('2'), '^', yy.number('2'))));
        });
        it('should left-associate divide and times', function () {
            expect(smarttype.parse('1/2*3'))
                .to.eql(yy.binop(yy.binop(yy.number('1'), '/', yy.number('2')), '*', yy.number(3)));
            expect(smarttype.parse('1*2/3'))
                .to.eql(yy.binop(yy.binop(yy.number('1'), '*', yy.number('2')), '/', yy.number(3)));
        });
        it('should left-associate plus and minus', function () {
            expect(smarttype.parse('1-2+3'))
                .to.eql(yy.binop(yy.binop(yy.number('1'), '-', yy.number('2')), '+', yy.number(3)));
            expect(smarttype.parse('1+2-3'))
                .to.eql(yy.binop(yy.binop(yy.number('1'), '+', yy.number('2')), '-', yy.number(3)));
        });
    });
    describe('[function]', function () {
        it('should allow nested functions', function () {
            expect(smarttype.parse('sinsinx'))
                .to.eql(yy.func('sin', yy.func('sin', yy.variable('x'))));
        });
        it('should allow superscripted functions', function () {
            expect(smarttype.parse('sin^-1x'))
                .to.eql(yy.func(yy.binop('sin', '^', yy.sign('-', yy.number('1'))), yy.variable('x')));
        });
        it('should allow subscripted functions', function () {
            expect(smarttype.parse('log_10x'))
                .to.eql(yy.func(yy.binop('log', '_', yy.number(10)), yy.variable('x')));
        });
        it('should allow negated function operands', function () {
            expect(smarttype.parse('sin-x'))
                .to.eql(yy.func('sin', yy.sign('-', yy.variable('x'))));
        });
        it('should allow double negated function operands', function () {
            expect(smarttype.parse('sin--x'))
                .to.eql(yy.func('sin', yy.sign('-', yy.sign('-', yy.variable('x')))));
        });
    });
    describe('[sub/superscripting]', function () {
        it('should handle mixed sub- and superscripting', function () {
            expect(smarttype.parse('x^x^x_x^2'))
                .to.eql(yy.binop(yy.binop(yy.binop(yy.variable('x'), '^',
                                                   yy.binop(yy.variable('x'), '^', yy.variable('x'))),
                                          '_', yy.variable('x')),
                                 '^', yy.number('2')));
        });
        it('should allow negative exponents', function () {
            expect(smarttype.parse('x^-2'))
                .to.eql(yy.binop(yy.variable('x'), '^', yy.sign('-', yy.number('2'))));
        });
        it.skip('should bind root higher than subscripts', function () {
            // don't know if we can fix this one
            expect(smarttype.parse('x_2rootx'))
                .to.eql(yy.binop(yy.variable('x'), '_', yy.binop(yy.number('2'), 'root', yy.variable('x'))));
        });
    });
    describe('[n\'th roots]', function () {
        it('should allow negative values in the exponent', function () {
            expect(smarttype.parse('-xroot2'))
                .to.eql(yy.binop(yy.sign('-', yy.variable('x')), 'root', yy.number('2')));
        });
        it('should allow functions within the exponent', function () {
            expect(smarttype.parse('cosxroot2'))
                .to.eql(yy.binop(yy.func('cos', yy.variable('x')), 'root', yy.number('2')));
        });
        it('show allow implicit multiplication of operands within functions in the exponent', function () {
            expect(smarttype.parse('cosxyroot2'))
                .to.eql(yy.binop(yy.func('cos', yy.binop(yy.variable('x'), '(*)', yy.variable('y'))), 'root', yy.number('2')));
        });
    });
    describe('[limits/product/sum]', function () {
        it('should allow nested limits/product/sums', function () {
            expect(smarttype.parse('lim_n->inf sum_i=1^n i'))
                .to.eql(yy.func(yy.binop('lim', '_', yy.binop(yy.variable('n'), '->', yy.symbol('inf'))), yy.func(yy.bigop('sum', yy.binop(yy.variable('i'), '=', yy.number('1')), yy.variable('n')), yy.variable('i'))));
        });
        it('should allow implicit multiplication of sums', function () {
            expect(smarttype.parse('2sum x'))
                .to.eql(yy.binop(yy.number('2'), '(*)', yy.func('sum', yy.variable('x'))));
        });
        it('should bind implicit multiplication higher than sum', function () {
            expect(smarttype.parse('sum xy'))
                .to.eql(yy.func('sum', yy.binop(yy.variable('x'), '(*)', yy.variable('y'))));
        });
    });
    describe('[spaces]', function () {
        it('should allow spaces before integrals', function () {
            expect(smarttype.parse(' int x dx'))
                .to.eql(yy.bracket('int', yy.variable('x'), 'dx', false));
        });
        it('should allow spaces before infinity', function () {
            expect(smarttype.parse(' inf'))
                .to.eql(yy.symbol('inf'));
        });
        it('should allow spaces after subscripted functions', function () {
            expect(smarttype.parse('log_10 x'))
                .to.eql(yy.func(yy.binop('log', '_', yy.number(10)), yy.variable('x')));
        });
        it('should allow spaces after symbols', function () {
            expect(smarttype.parse('alpha '))
                .to.eql(yy.symbol('alpha'));
        });
    });
    describe('[error handling]', function () {
        describe('[brackets]', function () {
            it('should handle incomplete brackets', function () {
                expect(smarttype.parse('()'))
                    .to.eql(yy.bracket('(', yy.symbol(null), ')', false));
                expect(smarttype.parse('||'))
                    .to.eql(yy.bracket('|', yy.symbol(null), '|', false));
            });
            it('should handle incomplete integral', function () {
                expect(smarttype.parse('int dx'))
                    .to.eql(yy.bracket('int', yy.symbol(null), 'dx', false));
            });
        });
        describe('[limits/product/sum]', function () {
            it('should handle incomplete sum', function () {
                expect(smarttype.parse('sum'))
                    .to.eql(yy.func('sum', yy.symbol(null)));
                expect(smarttype.parse('sum_0^1'))
                    .to.eql(yy.func(yy.bigop('sum', yy.number(0), yy.number(1)), yy.symbol(null)));
                expect(smarttype.parse('int sum_0^1'))
                    .to.eql(yy.bracket('int', yy.func(yy.bigop('sum', yy.number(0), yy.number(1)), yy.symbol(null)), undefined, false));
                expect(smarttype.parse('int sum_0^1 dx'))
                    .to.eql(yy.bracket('int', yy.func(yy.bigop('sum', yy.number(0), yy.number(1)), yy.symbol(null)), 'dx', false));
            });
            it.skip('should handle sum with incomplete binary operation', function () {
                expect(smarttype.parse('sum_0^1/2'))
                    .to.eql(yy.func(yy.bigop('sum', yy.number(0), yy.number(1), yy.binop(yy.symbol(null), '/', yy.number(2)))));
            });
        });
    });
});
