/* eslint-env node, mocha */
"use strict";

var smarttype = require('../lib/smarttype').parser;
var Simplifier = require('../lib/simplifier').Simplifier;
var expect = require("expect.js");

var oldyy = smarttype.yy;

var standard = new Simplifier({});
var autoparen = new Simplifier({autoparen: true});

describe('Simplifier', function () {
    afterEach(function () {
        smarttype.yy = oldyy;
    });
    describe('[parser connection]', function () {
        it('should receive atoms from the parser', function () {
            var yy = smarttype.yy = standard;
            expect(smarttype.parse('alpha')).to.eql(yy.symbol('alpha'));
            expect(smarttype.parse('1')).to.eql(yy.number('1'));
            expect(smarttype.parse('x')).to.eql(yy.variable('x'));
        });
        it('should receive binary operations from the parser', function () {
            var yy = smarttype.yy = standard;
            expect(smarttype.parse('x+y'))
                .to.eql(yy.binop(yy.variable('x'), '+', yy.variable('y')));
        });
        it('should receive big operations from the parser', function () {
            var yy = smarttype.yy = standard;
            expect(smarttype.parse('sum_x=0^inf1/x'))
                .to.eql(yy.func(yy.bigop('sum',
                                         yy.binop(yy.variable('x'), '=', yy.number('0')),
                                         yy.symbol('inf')),
                                yy.binop(yy.number('1'), '/', yy.variable('x'))));
        });
    });
    describe('[basic]', function () {
        it('should handle basic addition', function () {
            var yy = standard;
            expect(yy.binop(yy.variable('x'), '+', yy.variable('y')))
                .to.eql(new Simplifier.Addition({
                    operands: [new Simplifier.Variable({name: 'x'}),
                               new Simplifier.Variable({name: 'y'})],
                    signs: ['+']}));
            expect(yy.binop(yy.variable('x'), '-', yy.variable('y')))
                .to.eql(new Simplifier.Addition({
                    operands: [new Simplifier.Variable({name: 'x'}),
                               new Simplifier.Variable({name: 'y'})],
                    signs: ['-']}));
            expect(yy.binop(yy.variable('x'), '\\pm', yy.variable('y')))
                .to.eql(new Simplifier.Addition({
                    operands: [new Simplifier.Variable({name: 'x'}),
                               new Simplifier.Variable({name: 'y'})],
                    signs: ['\\pm']}));
            expect(yy.binop(yy.variable('x'), '\\pm', yy.variable('y')))
                .to.eql(new Simplifier.Addition({
                    operands: [new Simplifier.Variable({name: 'x'}),
                               new Simplifier.Variable({name: 'y'})],
                    signs: ['\\pm']}));
        });
        it('should handle basic multiplication', function () {
            var yy = standard;
            expect(yy.binop(yy.variable('x'), '*', yy.variable('y')))
                .to.eql(new Simplifier.Multiplication({
                    operands: [new Simplifier.Variable({name: 'x'}),
                               new Simplifier.Variable({name: 'y'})],
                    visible: [false]}));
        });
        it('should show multiplication when necessary', function () {
            var yy = standard;
            expect(yy.binop(yy.number('1'), '*', yy.number('2')))
                .to.eql(new Simplifier.Multiplication({
                    operands: [new Simplifier.Number({value: '1'}),
                               new Simplifier.Number({value: '2'})],
                    visible: [true]}));
            expect(yy.binop(yy.number('1'), '*', yy.binop(yy.number('2'), '^', yy.variable('x'))))
                .to.eql(new Simplifier.Multiplication({
                    operands: [new Simplifier.Number({value: '1'}),
                               new Simplifier.Binop({
                                   op: '^',
                                   left: new Simplifier.Number({value: '2'}),
                                   right: new Simplifier.Variable({name: 'x'})
                               })],
                    visible: [true]}));
        });
        it('should handle regular binary operations', function () {
            var yy = standard;
            expect(yy.binop(yy.variable('x'), '/', yy.variable('y')))
                .to.eql(new Simplifier.Binop({
                    op: '/',
                    left: new Simplifier.Variable({name: 'x'}),
                    right: new Simplifier.Variable({name: 'y'})}));
        });
    });
    describe('[matrix]', function () {
        it('should handle row matrices', function () {
            var yy = standard;
            expect(yy.binop(yy.binop(yy.variable('x'), '&', yy.variable('y')), '&', yy.variable('z')))
                .to.eql(new Simplifier.Matrix({
                    elements: [[new Simplifier.Variable({name: 'x'}),
                                new Simplifier.Variable({name: 'y'}),
                                new Simplifier.Variable({name: 'z'})]],
                    width: 3,
                    height: 1
                }));
        });
        it('should handle column matrices', function () {
            var yy = standard;
            expect(yy.binop(yy.binop(yy.variable('x'), '\\', yy.variable('y')), '\\', yy.variable('z')))
                .to.eql(new Simplifier.Matrix({
                    elements: [[new Simplifier.Variable({name: 'x'})],
                               [new Simplifier.Variable({name: 'y'})],
                               [new Simplifier.Variable({name: 'z'})]],
                    width: 1,
                    height: 3
                }));
        });
        it('should handle 2D matrices', function () {
            var yy = standard;
            expect(yy.binop(yy.binop(yy.binop(yy.variable('x'), '&', yy.variable('y')), '&', yy.variable('z')), '\\',
                            yy.binop(yy.binop(yy.variable('a'), '&', yy.variable('b')), '&', yy.variable('c'))))
                .to.eql(new Simplifier.Matrix({
                    elements: [[new Simplifier.Variable({name: 'x'}),
                                new Simplifier.Variable({name: 'y'}),
                                new Simplifier.Variable({name: 'z'})],
                               [new Simplifier.Variable({name: 'a'}),
                                new Simplifier.Variable({name: 'b'}),
                                new Simplifier.Variable({name: 'c'})]],
                    width: 3,
                    height: 2
                }));
        });
    });
    describe('[simplification]', function () {
        describe('[binary operations]', function () {
            it('should group multiple addition', function () {
                var yy = standard;
                expect(yy.binop(yy.binop(yy.variable('x'), '+', yy.variable('y')), '+', yy.variable('z')))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'}),
                                   new Simplifier.Variable({name: 'z'})],
                        signs: ['+', '+']
                    }));
                expect(yy.binop(yy.binop(yy.binop(yy.variable('w'), '+', yy.variable('x')), '-', yy.variable('y')), '+', yy.variable('z')))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'w'}),
                                   new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'}),
                                   new Simplifier.Variable({name: 'z'})],
                        signs: ['+', '-', '+']
                    }));
            });
            it('should simplify bracketed addition', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('x'), '+', yy.variable('y')), ')', false), '+', yy.variable('z')))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'}),
                                   new Simplifier.Variable({name: 'z'})],
                        signs: ['+', '+']
                    }));
                expect(yy.binop(yy.bracket('[', yy.binop(yy.variable('x'), '+', yy.variable('y')), ']', false), '+', yy.variable('z')))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'}),
                                   new Simplifier.Variable({name: 'z'})],
                        signs: ['+', '+']
                    }));
                expect(yy.binop(yy.bracket('{', yy.binop(yy.variable('x'), '+', yy.variable('y')), '}', false), '+', yy.variable('z')))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'}),
                                   new Simplifier.Variable({name: 'z'})],
                        signs: ['+', '+']
                    }));
                expect(yy.binop(yy.variable('x'), '+', yy.bracket('(', yy.binop(yy.variable('y'), '+', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'}),
                                   new Simplifier.Variable({name: 'z'})],
                        signs: ['+', '+']
                    }));
            });
            it('should not simplify bracketed operands on the rhs of a subtration ', function () {
                var yy = autoparen;
                expect(yy.binop(yy.variable('x'), '-', yy.bracket('(', yy.binop(yy.variable('y'), '+', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Bracket({
                                       lbracket: '(',
                                       value: new Simplifier.Addition({
                                           operands: [new Simplifier.Variable({name: 'y'}),
                                                      new Simplifier.Variable({name: 'z'})],
                                           signs: ['+']
                                       }),
                                       rbracket: ')',
                                       hidden: false
                                   })],
                        signs: ['-']
                    }));
            });
            it('should lift higher precedence bracketed expressions out in addition expressions', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '*', yy.variable('x')), ')', false),
                                '+', yy.bracket('(', yy.binop(yy.variable('y'), '*', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [
                            new Simplifier.Multiplication({
                                operands: [new Simplifier.Variable({name: 'w'}),
                                           new Simplifier.Variable({name: 'x'})],
                                visible: [false]
                            }),
                            new Simplifier.Multiplication({
                                operands: [new Simplifier.Variable({name: 'y'}),
                                           new Simplifier.Variable({name: 'z'})],
                                visible: [false]
                            })],
                        signs: ['+']
                    }));
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '^', yy.variable('x')), ')', false),
                                '+', yy.bracket('(', yy.binop(yy.variable('y'), '^', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [
                            new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'w'}),
                                op: '^',
                                right: new Simplifier.Variable({name: 'x'})
                            }),
                            new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'y'}),
                                op: '^',
                                right: new Simplifier.Variable({name: 'z'})
                            })],
                        signs: ['+']
                    }));
                expect(yy.binop(yy.bracket('(', yy.func('sum', yy.variable('x')), ')', false),
                                '+', yy.bracket('(', yy.func('sum', yy.variable('y')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [
                            new Simplifier.Func({
                                name: 'sum',
                                value: new Simplifier.Variable({name: 'x'})
                            }),
                            new Simplifier.Func({
                                name: 'sum',
                                value: new Simplifier.Variable({name: 'y'})
                            })],
                        signs: ['+']
                    }));
                expect(yy.binop(yy.bracket('(', yy.variable('x'), ')', false),
                                '+', yy.bracket('(', yy.variable('y'), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'})],
                        signs: ['+']
                    }));
            });
            it('should lift higher precedence bracketed expressions out in subtraction expressions', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '*', yy.variable('x')), ')', false),
                                '-', yy.bracket('(', yy.binop(yy.variable('y'), '*', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [
                            new Simplifier.Multiplication({
                                operands: [new Simplifier.Variable({name: 'w'}),
                                           new Simplifier.Variable({name: 'x'})],
                                visible: [false]
                            }),
                            new Simplifier.Multiplication({
                                operands: [new Simplifier.Variable({name: 'y'}),
                                           new Simplifier.Variable({name: 'z'})],
                                visible: [false]
                            })],
                        signs: ['-']
                    }));
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '^', yy.variable('x')), ')', false),
                                '-', yy.bracket('(', yy.binop(yy.variable('y'), '^', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [
                            new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'w'}),
                                op: '^',
                                right: new Simplifier.Variable({name: 'x'})
                            }),
                            new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'y'}),
                                op: '^',
                                right: new Simplifier.Variable({name: 'z'})
                            })],
                        signs: ['-']
                    }));
                expect(yy.binop(yy.bracket('(', yy.func('sum', yy.variable('x')), ')', false),
                                '-', yy.bracket('(', yy.func('sum', yy.variable('y')), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [
                            new Simplifier.Func({
                                name: 'sum',
                                value: new Simplifier.Variable({name: 'x'})
                            }),
                            new Simplifier.Func({
                                name: 'sum',
                                value: new Simplifier.Variable({name: 'y'})
                            })],
                        signs: ['-']
                    }));
                expect(yy.binop(yy.bracket('(', yy.variable('x'), ')', false),
                                '-', yy.bracket('(', yy.variable('y'), ')', false)))
                    .to.eql(new Simplifier.Addition({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'})],
                        signs: ['-']
                    }));
            });
            it('should simplify bracketed multiplication', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '*', yy.variable('x')), ')', false),
                                '*', yy.bracket('(', yy.binop(yy.variable('y'), '*', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Multiplication({
                        operands: [ new Simplifier.Variable({name: 'w'}),
                                    new Simplifier.Variable({name: 'x'}),
                                    new Simplifier.Variable({name: 'y'}),
                                    new Simplifier.Variable({name: 'z'})],
                        visible: [false, false, false]
                    }));
            });
            it('should lift higher precedence bracketed expressions out in multiplication expressions', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '^', yy.variable('x')), ')', false),
                                '*', yy.bracket('(', yy.binop(yy.variable('y'), '^', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Multiplication({
                        operands: [
                            new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'w'}),
                                op: '^',
                                right: new Simplifier.Variable({name: 'x'})
                            }),
                            new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'y'}),
                                op: '^',
                                right: new Simplifier.Variable({name: 'z'})
                            })],
                        visible: [false]
                    }));
                expect(yy.binop(yy.bracket('(', yy.func('sum', yy.variable('x')), ')', false),
                                '*', yy.bracket('(', yy.func('sum', yy.variable('y')), ')', false)))
                    .to.eql(new Simplifier.Multiplication({
                        operands: [
                            new Simplifier.Func({
                                name: 'sum',
                                value: new Simplifier.Variable({name: 'x'})
                            }),
                            new Simplifier.Func({
                                name: 'sum',
                                value: new Simplifier.Variable({name: 'y'})
                            })],
                        visible: [false]
                    }));
                expect(yy.binop(yy.bracket('(', yy.variable('x'), ')', false),
                                '*', yy.bracket('(', yy.variable('y'), ')', false)))
                    .to.eql(new Simplifier.Multiplication({
                        operands: [new Simplifier.Variable({name: 'x'}),
                                   new Simplifier.Variable({name: 'y'})],
                        visible: [false]
                    }));
            });
            it('should un-bracket numerators and denominators', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '+', yy.variable('x')), ')', false),
                                '/', yy.bracket('(', yy.binop(yy.variable('y'), '+', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Binop({
                        left: new Simplifier.Addition({
                            operands: [new Simplifier.Variable({name: 'w'}),
                                       new Simplifier.Variable({name: 'x'})],
                            signs: ['+']
                        }),
                        op: '/',
                        right: new Simplifier.Addition({
                            operands: [new Simplifier.Variable({name: 'y'}),
                                       new Simplifier.Variable({name: 'z'})],
                            signs: ['+']
                        })
                    }));
            });
            it('should un-bracket n\'th roots', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('w'), '+', yy.variable('x')), ')', false),
                                'root', yy.bracket('(', yy.binop(yy.variable('y'), '+', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Binop({
                        left: new Simplifier.Addition({
                            operands: [new Simplifier.Variable({name: 'w'}),
                                       new Simplifier.Variable({name: 'x'})],
                            signs: ['+']
                        }),
                        op: 'root',
                        right: new Simplifier.Addition({
                            operands: [new Simplifier.Variable({name: 'y'}),
                                       new Simplifier.Variable({name: 'z'})],
                            signs: ['+']
                        })
                    }));
            });
            it('should un-bracket rhs of sub/super-scripts', function () {
                var yy = autoparen;
                expect(yy.binop(yy.variable('x'),
                                '^', yy.bracket('(', yy.binop(yy.variable('y'), '+', yy.variable('z')), ')', false)))
                    .to.eql(new Simplifier.Binop({
                        left: new Simplifier.Variable({name: 'x'}),
                        op: '^',
                        right: new Simplifier.Addition({
                            operands: [new Simplifier.Variable({name: 'y'}),
                                       new Simplifier.Variable({name: 'z'})],
                            signs: ['+']
                        })
                    }));
            });
            it('should not un-bracket lhs of sub/super-scripts', function () {
                var yy = autoparen;
                expect(yy.binop(yy.bracket('(', yy.binop(yy.variable('x'), '+', yy.variable('y')), ')', false),
                                '^', yy.variable('z')))
                    .to.eql(new Simplifier.Binop({
                        left: new Simplifier.Bracket({
                            lbracket: '(',
                            value: new Simplifier.Addition({
                                operands: [new Simplifier.Variable({name: 'x'}),
                                           new Simplifier.Variable({name: 'y'})],
                                signs: ['+']
                            }),
                            rbracket: ')',
                            hidden: false
                        }),
                        op: '^',
                        right: new Simplifier.Variable({name: 'z'})
                    }));
            });
        });
        describe('[functions]', function () {
            it('should un-bracket square roots', function () {
                var yy = autoparen;
                expect(yy.func('sqrt', yy.bracket('(', yy.binop(yy.variable('x'), '+', yy.variable('y')), ')', false)))
                    .to.eql(new Simplifier.Func({
                        name: 'sqrt',
                        value: new Simplifier.Addition({
                            operands: [new Simplifier.Variable({name: 'x'}),
                                       new Simplifier.Variable({name: 'y'})],
                            signs: ['+']
                        })
                    }));
            });
            it('should un-bracket sub/super-scripts of big ops', function () {
                var yy = autoparen;
                expect(yy.bigop('sum', yy.bracket('(', yy.binop(yy.variable('x'), '=', yy.number('0')), ')', false),
                                yy.bracket('(', yy.binop(yy.variable('n'), '-', yy.number('1')), ')', false)))
                    .to.eql(new Simplifier.Bigop({
                        op: 'sum',
                        sub: new Simplifier.Binop({
                            left: new Simplifier.Variable({name: 'x'}),
                            op: '=',
                            right: new Simplifier.Number({value: '0'})
                        }),
                        sup: new Simplifier.Addition({
                            operands: [new Simplifier.Variable({name: 'n'}),
                                       new Simplifier.Number({value: '1'})],
                            signs: ['-']
                        })
                    }));
            });
            it('should un-bracket atomic operands of big ops', function () {
                var yy = autoparen;
                expect(yy.func('sum', yy.bracket('(', yy.variable('x'), ')', false)))
                    .to.eql(new Simplifier.Func({
                        name: 'sum',
                        value: new Simplifier.Variable({name: 'x'})
                    }));
                expect(yy.func(yy.bigop('sum', yy.binop(yy.variable('x'), '=', yy.number('0')), yy.variable('n')),
                               yy.bracket('(', yy.variable('x'), ')', false)))
                    .to.eql(new Simplifier.Func({
                        name: new Simplifier.Bigop({
                            op: 'sum',
                            sub: new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'x'}),
                                op: '=',
                                right: new Simplifier.Number({value: '0'})
                            }),
                            sup: new Simplifier.Variable({name: 'n'})
                        }),
                        value: new Simplifier.Variable({name: 'x'})
                    }));
            });
            it('should not change other operands of big ops', function () {
                var yy = autoparen;
                expect(yy.func('sum', yy.bracket('(', yy.binop(yy.variable('x'), '+', yy.number('1')), ')', false)))
                    .to.eql(new Simplifier.Func({
                        name: 'sum',
                        value: new Simplifier.Bracket({
                            lbracket: '(',
                            value: new Simplifier.Addition({
                                operands: [new Simplifier.Variable({name: 'x'}),
                                           new Simplifier.Number({value: '1'})],
                                signs: ['+']
                            }),
                            rbracket: ')',
                            hidden: false
                        })
                    }));
                expect(yy.func(yy.bigop('sum', yy.binop(yy.variable('x'), '=', yy.number('0')), yy.variable('n')),
                               yy.bracket('(', yy.binop(yy.variable('x'), '+', yy.number('1')), ')', false)))
                    .to.eql(new Simplifier.Func({
                        name: new Simplifier.Bigop({
                            op: 'sum',
                            sub: new Simplifier.Binop({
                                left: new Simplifier.Variable({name: 'x'}),
                                op: '=',
                                right: new Simplifier.Number({value: '0'})
                            }),
                            sup: new Simplifier.Variable({name: 'n'})
                        }),
                        value: new Simplifier.Bracket({
                            lbracket: '(',
                            value: new Simplifier.Addition({
                                operands: [new Simplifier.Variable({name: 'x'}),
                                           new Simplifier.Number({value: '1'})],
                                signs: ['+']
                            }),
                            rbracket: ')',
                            hidden: false
                        })
                    }));
            });
            it('should not bracket absolute values or braces', function () {
                var yy = autoparen;
                expect(yy.func('sin', yy.bracket('|', yy.variable('x'), '|', false)))
                    .to.eql(new Simplifier.Func({
                        name: 'sin',
                        value: new Simplifier.Bracket({
                            lbracket: '|',
                            value: new Simplifier.Variable({name: 'x'}),
                            rbracket: '|',
                            hidden: false
                        })
                    }));
                expect(yy.func('sin', yy.bracket('{', yy.variable('x'), '}', false)))
                    .to.eql(new Simplifier.Func({
                        name: 'sin',
                        value: new Simplifier.Bracket({
                            lbracket: '{',
                            value: new Simplifier.Variable({name: 'x'}),
                            rbracket: '}',
                            hidden: false
                        })
                    }));
            });
            it('should re-wrap other brackets', function () {
                var yy = autoparen;
                expect(yy.func('sin', yy.bracket('{', yy.variable('x'), '|', false)))
                    .to.eql(new Simplifier.Func({
                        name: 'sin',
                        value: new Simplifier.Bracket({
                            lbracket: '(',
                            value: new Simplifier.Variable({name: 'x'}),
                            rbracket: ')',
                            hidden: false
                        })
                    }));
                expect(yy.func('sin', yy.bracket('[', yy.variable('x'), ']', false)))
                    .to.eql(new Simplifier.Func({
                        name: 'sin',
                        value: new Simplifier.Bracket({
                            lbracket: '(',
                            value: new Simplifier.Variable({name: 'x'}),
                            rbracket: ')',
                            hidden: false
                        })
                    }));
            });
            it('should wrap unwrapped expressions', function () {
                var yy = autoparen;
                expect(yy.func('sin', yy.variable('x')))
                    .to.eql(new Simplifier.Func({
                        name: 'sin',
                        value: new Simplifier.Bracket({
                            lbracket: '(',
                            value: new Simplifier.Variable({name: 'x'}),
                            rbracket: ')',
                            hidden: false
                        })
                    }));
            });
        });
        describe('[brackets]', function () {
            it('should strip nested parentheses', function () {
                var yy = autoparen;
                expect(yy.bracket('(', yy.bracket('(', yy.variable('x'), ')', false), ')', false))
                    .to.eql(new Simplifier.Bracket({
                        lbracket: '(',
                        value: new Simplifier.Variable({name: 'x'}),
                        rbracket: ')',
                        hidden: false
                    }));
            });
            it('should strip parentheses inside integrals', function () {
                var yy = autoparen;
                expect(yy.bracket('int', yy.bracket('(', yy.variable('x'), ')', false), 'dx', false))
                    .to.eql(new Simplifier.Bracket({
                        lbracket: 'int',
                        value: new Simplifier.Variable({name: 'x'}),
                        rbracket: 'dx',
                        hidden: false
                    }));
            });
        });
    });
});
